# enginegrid.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for listing details of chess engines enabled for chess database.
"""

import tkinter
from urllib.parse import urlsplit, parse_qs

from solentware_misc.workarounds import dialogues

from solentware_grid.datagrid import DataGrid

from ..core.chessrecord import ChessDBrecordEngine
from .enginerow import ChessDBrowEngine
from .chessexception import ChessException
from .eventspec import EventSpec, DummyEvent
from .display import Display


class EngineListGrid(ChessException, DataGrid, Display):

    """A DataGrid for lists of chess engine definition.

    Subclasses provide navigation and extra methods appropriate to list use.

    """

    def __init__(self, parent):
        '''Extend with link to user interface object.

        parent - see superclass

        '''
        super(EngineListGrid, self).__init__(parent=parent)
        self.gcanvas.configure(takefocus=tkinter.FALSE)
        self.data.configure(takefocus=tkinter.FALSE)
        self.frame.configure(takefocus=tkinter.FALSE)
        self.hsbar.configure(takefocus=tkinter.FALSE)
        self.vsbar.configure(takefocus=tkinter.FALSE)
        
    def set_properties(self, key, dodefaultaction=True):
        """Return True if chess engine definition properties set or False."""
        if super(EngineListGrid, self).set_properties(
            key, dodefaultaction=False):
            return True
        if dodefaultaction:
            self.objects[key].set_background_normal(self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for chess engine definition key or None."""
        row = super(EngineListGrid, self).set_row(
            key, dodefaultaction=False, **kargs)
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        else:
            return None

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordEngine()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue))
        self.create_delete_dialog(
            self.objects[key],
            oldobject,
            modal,
            title='Delete Engine Definition')

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordEngine(),
            ChessDBrecordEngine(),
            False,
            modal,
            title='Edit Engine Definition')

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordEngine(),
            ChessDBrecordEngine(),
            True,
            modal,
            title='Edit Engine Definition')

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        newobject = ChessDBrecordEngine()
        instance = self.datasource.new_row()
        instance.srvalue = instance.value.pack_value()
        self.create_edit_dialog(
            instance,
            newobject,
            None,
            False,
            modal,
            title='New Engine Definition')

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordEngine()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue))
        self.create_show_dialog(
            self.objects[key],
            oldobject,
            modal,
            title='Show Engine Definition')
        
    def create_edit_dialog(
        self, instance, newobject, oldobject, showinitial, modal, title=''):
        """Extend to do chess initialization"""
        for x in (newobject, oldobject):
            if x:
                x.load_record((instance.key.pack(), instance.srvalue))
        super(EngineListGrid, self).create_edit_dialog(
            instance, newobject, oldobject, showinitial, modal, title=title)

    def fill_view(
        self,
        currentkey=None,
        down=True,
        topstart=True,
        exclude=True,
        ):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(EngineListGrid, self).fill_view(
                currentkey=currentkey,
                down=down,
                topstart=topstart,
                exclude=exclude,
                )

    def load_new_index(self):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(EngineListGrid, self).load_new_index()

    def load_new_partial_key(self, key):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(EngineListGrid, self).load_new_partial_key(key)

    def on_configure_canvas(self, event=None):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(EngineListGrid, self).on_configure_canvas(event=event)

    def on_data_change(self, instance):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(EngineListGrid, self).on_data_change(instance)

            # Hack to display newly inserted record.
            # Acceptable because there will likely be two or three records at
            # most in the engine grid.
            self.fill_view_from_top()

    def set_focus(self):
        """Give focus to this widget."""
        self.frame.focus_set()

    def is_payload_available(self):
        """Return True if grid is connected to a database."""
        ds = self.get_data_source()
        if ds is None:
            return False
        if ds.get_database() is None:

            # Avoid exception scrolling visible grid not connected to database.
            # Make still just be hack to cope with user interface activity
            # while importing data.
            self.clear_grid_keys()

            return False
        return True

    def bind_for_widget_without_focus(self):
        """Return True if this item has the focus about to be lost."""
        if self.get_frame().focus_displayof() != self.get_frame():
            return False

        # Nothing to do on losing focus.
        return True
        
    def get_top_widget(self):
        """Return topmost widget for game display.

        The topmost widget is put in a container widget in some way

        """
        # Superclass DataGrid.get_frame() method returns the relevant widget.
        # Name, get_top_widget, is compatible with Game and Partial names.
        return self.get_frame()
        

class EngineGrid(EngineListGrid):

    """Customized EngineListGrid for list of enabled chess engines.
    """

    def __init__(self, ui):
        '''Extend with definition and bindings for selection rules on grid.

        ui - container for user interface widgets and methods.

        '''
        super(EngineGrid, self).__init__(ui.show_engines_toplevel)
        self.ui = ui
        self.make_header(ChessDBrowEngine.header_specification)
        self.__bind_on()
        for function, accelerator in (
            (self.run_engine,
             EventSpec.engine_grid_run),
            ):
            self.menupopup.insert_command(
                0,
                label=accelerator[1],
                command=self.try_command(function, self.menupopup),
                accelerator=accelerator[2])

    def bind_off(self):
        """Disable all bindings."""
        super(EngineGrid, self).bind_off()
        for function, accelerator in (
            ('',
             EventSpec.engine_grid_run),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(accelerator[0], function)

    def bind_on(self):
        """Enable all bindings."""
        super(EngineGrid, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        for function, accelerator in (
            (self.run_engine,
             EventSpec.engine_grid_run),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(accelerator[0], function)
        
    def on_partial_change(self, instance):
        """Delegate to superclass if database is open otherwise do nothing."""

        # may turn out to be just to catch datasource is None
        if self.get_data_source() is None:
            return
        super(EngineGrid, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display selection rule name."""

    def is_visible(self):
        """Return True if list of selection rules is displayed."""
        #return str(self.get_frame()) in self.ui.selection_rules_pw.panes()
        return True

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.
        
        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.
        
        """
        try:
            super().set_selection(key)
        except KeyError:
            dialogues.showinfo(
                title='Insert Engine Definition Workaround',
                message=''.join(
                    ('All records have same name on this display.\n\nThe new ',
                     'record has been inserted but you need to Hide, and then ',
                     'Show, the display to see the record in the list.',
                     )))

    def run_engine(self, event=None):
        """Run chess engine."""
        self.launch_chess_engine(self.pointer_popup_selection)
        #self.move_selection_to_popup_selection()

    def launch_chess_engine(self, key, modal=True):
        """Launch a chess engine."""
        oldobject = ChessDBrecordEngine()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue))
        definition = oldobject.value

        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under Wine on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        #if self.ui.uci.uci_drivers_reply is None:
        #    dialogues.showinfo(
        #        'Chesstab Restriction',
        #        ' '.join(('Starting an UCI chess engine is not allowed because',
        #                  'an interface cannot be created:',
        #                  'this is expected if running under Wine.')))
        #    return

        url = urlsplit(definition.get_engine_command_text())
        try:
            url.port
        except ValueError as exc:
            dialogues.showerror(
                title='Run Engine',
                message=''.join(('The port in the chess engine definition is ',
                                 'invalid.\n\n',
                                 'The reported error for the port is:\n\n',
                                 str(exc),
                                 )))
            return
        if not definition.get_engine_command_text():
            dialogues.showinfo(
                title='Run Engine',
                message=''.join((
                    'The engine definition does not have a command to ',
                    'run chess engine.',
                    )))
            return
        elif not (url.port or url.hostname):
            if not definition.is_run_engine_command():
                dialogues.showinfo(
                    title='Run Engine',
                    message=''.join((
                        'The engine definition command to run a chess engine ',
                        'does not name a file.',
                        )))
                return
        if url.hostname or url.port:
            if url.path and url.query:
                dialogues.showerror(
                    title='Run Engine',
                    message=''.join(
                        ('Engine must be query with hostname or port.\n\n',
                         "Path is: '", url.path, "'.\n\n",
                         "Query is: '", url.query, "'.\n",
                         )))
                return
            elif url.path:
                dialogues.showerror(
                    title='Run Engine',
                    message=''.join(
                        ('Engine must be query with hostname or port.\n\n',
                         "Path is: '", url.path, "'.\n",
                         )))
                return
            elif not url.query:
                dialogues.showerror(
                    title='Run Engine',
                    message='Engine must be query with hostname or port.\n\n')
                return
            else:
                try:
                    query = parse_qs(url.query, strict_parsing=True)
                except ValueError as exc:
                    dialogues.showerror(
                        title='Run Engine',
                        message=''.join(
                            ("Problem in chess engine specification.  ",
                             "The reported error is:\n\n'",
                             str(exc), "'.\n",
                             )))
                    return
                if len(query) > 1:
                    dialogues.showerror(
                        title='Run Engine',
                        message=''.join(
                            ("Engine must be single 'key=value' or ",
                             "'value'.\n\n",
                             "Query is: '", url.query, "'.\n",
                             )))
                    return
                elif len(query) == 1:
                    for k, v in query.items():
                        if k != 'name':
                            dialogues.showerror(
                                title='Run Engine',
                                message=''.join(
                                    ("Engine must be single 'key=value' or ",
                                     "'value'.\n\n",
                                     "Query is: '", url.query, "'\n\nand use ",
                                     "'name' as key.\n",
                                     )))
                            return
                        elif len(v) > 1:
                            dialogues.showerror(
                                title='Run Engine',
                                message=''.join(
                                    ("Engine must be single 'key=value' or ",
                                     "'value'.\n\n",
                                     "Query is: '", url.query, "' with more ",
                                     "than one 'value'\n",
                                     )))
                            return
        elif url.path and url.query:
            dialogues.showerror(
                title='Run Engine',
                message=''.join(
                    ('Engine must be path without hostname or port.\n\n',
                     "Path is: '", url.path, "'.\n\n",
                     "Query is: '", url.query, "'.\n",
                     )))
            return
        elif url.query:
            dialogues.showerror(
                title='Run Engine',
                message=''.join(
                    ('Engine must be path without hostname or port.\n\n',
                     "Query is: '", url.query, "'.\n",
                     )))
            return
        elif not url.path:
            dialogues.showerror(
                title='Run Engine',
                message='Engine must be path without hostname or port.\n')
            return
        self.ui.run_engine(definition.get_engine_command_text())



# gamedbedit.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise edit dialogue to edit or insert chess game record.
"""

import tkinter

from solentware_grid.gui.dataedit import DataEdit

from solentware_misc.workarounds import dialogues

from pgn_read.core.constants import TAG_WHITE, TAG_BLACK

from .chessexception import ChessException
from .gamedisplay import DialogueGameDisplay, DialogueGameEdit
from .constants import EMPTY_SEVEN_TAG_ROSTER


class ChessDBeditGame(ChessException, DataEdit):
    
    """Dialog to edit a game on, or insert a game into, database.

    The game is in it's own Toplevel widget and playing through the game does
    not change the list of games, in the main widget, matching the current
    position on the board.

    """

    def __init__(self, newobject, parent, oldobject, showinitial=True, ui=None):
        """Extend and create dialogue widget to edit or insert chess game."""
        if oldobject:
            tags = oldobject.value.collected_game[1]
            try:
                title = '  '.join((
                    'Edit Game:',
                    ' - '.join((
                        tags[TAG_WHITE],
                        tags[TAG_BLACK])),
                    ))
            except TypeError:
                title = 'Edit Game - names unknown or invalid'
            except KeyError:
                title = 'Edit Game - names unknown or invalid'
        else:
            title = 'Insert Game'
            showinitial = False
        if showinitial:
            showinitial = DialogueGameDisplay(master=parent, ui=ui)
            showinitial.set_position_analysis_data_source()
            if ui is not None:
                ui.games_and_repertoires_in_toplevels.add(showinitial)
            showinitial.pgn.get_first_game(newobject.get_srvalue())
            showinitial.set_game()
        newview = DialogueGameEdit(master=parent, ui=ui)
        newview.set_position_analysis_data_source()
        if ui is not None:
            ui.games_and_repertoires_in_toplevels.add(newview)
        if oldobject:
            newview.pgn.get_first_game(newobject.get_srvalue())
        else:
            newview.pgn.get_first_game(''.join((EMPTY_SEVEN_TAG_ROSTER, '*')))
        newview.set_game()
        super(ChessDBeditGame, self).__init__(
            newobject,
            parent,
            oldobject,
            newview,
            title,
            oldview=showinitial,
            )

        # Bind only to newview.score and newview.analysis.score because these
        # alone takes focus.
        self.bind_buttons_to_widget(newview.score)
        self.bind_buttons_to_widget(newview.analysis.score)

        self.ui = ui

    def dialog_ok(self):
        """Update record and return update action response (True for updated).

        Check that database is open and is same one as update action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text='Cannot update because not connected to a database')
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        text = self.newview.get_score_error_escapes_removed()
        self.newobject.value.load(repr(text))
        if not self.newobject.value.is_pgn_valid():
            if tkinter.messagebox.YES != dialogues.askquestion(
                title='Edit Game',
                message=''.join(
                    ('The edited game score contains at least one illegal ',
                     'move in PGN.\n\nPlease re-confirm request to edit game.',
                     ))):
                return False
            self.newobject.value.set_game_source('Editor')
        return super(ChessDBeditGame, self).dialog_ok()
        
    def put(self):
        """Mark partial position records for recalculation and return key"""
        self.datasource.dbhome.mark_partial_positions_to_be_recalculated()
        super(ChessDBeditGame, self).put()

    def edit(self):
        """Mark partial position records for recalculation and return key"""
        self.datasource.dbhome.mark_partial_positions_to_be_recalculated()
        super(ChessDBeditGame, self).edit()

    def tidy_on_destroy(self):
        """Clear up after dialogue destruction."""
        self.ui.games_and_repertoires_in_toplevels.discard(self.oldview)
        self.ui.games_and_repertoires_in_toplevels.discard(self.newview)
        self.ui.base_games.selection.clear()



# chessdb.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database using Berkeley DB.

The Berkeley DB interface has significantly worse performance than DPT when
doing multi-index searches.  However it is retained since DPT became available
because tracking down problems in the chess logic using IDLE can be easier
in the *nix environment.
"""

import os

from bsddb3.db import (
    DB_BTREE,
    DB_HASH,
    DB_RECNO,
    DB_DUPSORT,
    DB_DUP,
    DB_CREATE,
    DB_RECOVER,
    DB_INIT_MPOOL,
    DB_INIT_LOCK,
    DB_INIT_LOG,
    DB_INIT_TXN,
    DB_PRIVATE,
    )

from solentware_base import bsddb3_database
from solentware_base.core.constants import (
    SUBFILE_DELIMITER,
    EXISTENCE_BITMAP_SUFFIX,
    SEGMENT_SUFFIX,
    SECONDARY,
    )

from ..core.filespec import (
    FileSpec,
    DB_ENVIRONMENT_GIGABYTES,
    DB_ENVIRONMENT_BYTES,
    DB_ENVIRONMENT_MAXLOCKS,
    )
from ..basecore import database


class ChessDatabase(database.Database, bsddb3_database.Database):

    """Provide access to a database of games of chess.
    """

    _deferred_update_process = os.path.join(
        os.path.basename(os.path.dirname(__file__)), 'runchessdbdu.py')

    def __init__(self, DBfile, **kargs):
        """Define chess database.

        **kargs
        Arguments are passed through to superclass __init__.
        
        """
        dbnames = FileSpec(**kargs)

        environment = {'flags': (DB_CREATE |
                                 DB_RECOVER |
                                 DB_INIT_MPOOL |
                                 DB_INIT_LOCK |
                                 DB_INIT_LOG |
                                 DB_INIT_TXN |
                                 DB_PRIVATE),
                       'gbytes': DB_ENVIRONMENT_GIGABYTES,
                       'bytes': DB_ENVIRONMENT_BYTES,
                       'maxlocks': DB_ENVIRONMENT_MAXLOCKS,
                       }

        super().__init__(
            dbnames,
            DBfile,
            environment)

    def delete_database(self):
        """Close and delete the open chess database."""
        super().delete_database((self.database_file,
                                 self.dbenv.get_lg_dir().decode()))



# cqledit.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Edit a Chess Query Language (ChessQL) statement.

The CQLEdit class displays a ChessQL statement and allows editing.

This class has the cql.CQL class as a superclass.

This class does not allow deletion of ChessQL statements from a database.

An instance of these classes fits into the user interface in two ways: as an
item in a panedwindow of the main widget, or as the only item in a new toplevel
widget.

"""

from .cql import CQL


class CQLEdit(CQL):
    
    """Display a ChessQL statement with editing allowed.
    """

    # True means ChessQL statement can be edited
    _is_cql_query_editable = True

    def __init__(self, **ka):
        """Extend ChessQL statement widget as editor."""
        super().__init__(**ka)
        self.bind_pointer()

    def bind_pointer(self):
        """Set pointer button-1 binding."""
        for sequence, function in (
            ('<Control-ButtonPress-1>', ''),
            ('<ButtonPress-1>', ''), # go_to_token
            ('<ButtonPress-3>', ''),
            ):
            self.score.bind(sequence, function)

    def disable_keyboard(self):
        """Override and do nothing."""



# querydbedit.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise edit dialogue to edit or insert game selection rule record.
"""

import tkinter.messagebox

from solentware_misc.workarounds import dialogues

from solentware_grid.gui.dataedit import DataEdit

from .chessexception import ChessException
from .querydisplay import DialogueQueryDisplay, DialogueQueryEdit


class ChessDBeditQuery(ChessException, DataEdit):
    """Dialog to edit a game selection rule on, or insert one into, database.

    The game selection rule is in it's own Toplevel widget.

    """

    def __init__(self, newobject, parent, oldobject, showinitial=True, ui=None):
        """Extend and create dialogue to edit or insert selection rule."""
        if oldobject:
            title = ':  '.join((
                'Edit Selection Rule',
                oldobject.value._description_string))
        else:
            title = 'Insert Selection Rule'
            showinitial = False
        self.__title = title.split(':')[0]
        if showinitial:
            showinitial = DialogueQueryDisplay(master=parent, ui=ui)
            if ui is not None:
                ui.selections_in_toplevels.add(showinitial)
                showinitial.query_statement.set_database(
                    ui.base_games.datasource.dbhome)
                showinitial.query_statement.dbset = (
                    ui.base_games.datasource.dbset)
            showinitial.query_statement.process_query_statement(
                oldobject.get_srvalue())
            showinitial.set_query_statement()
        newview = DialogueQueryEdit(master=parent, ui=ui)
        if ui is not None:
            ui.selections_in_toplevels.add(newview)
            newview.query_statement.set_database(
                ui.base_games.datasource.dbhome)
            newview.query_statement.dbset = ui.base_games.datasource.dbset
        newview.query_statement.process_query_statement(newobject.get_srvalue())
        newview.set_query_statement()
        super(ChessDBeditQuery, self).__init__(
            newobject,
            parent,
            oldobject,
            newview,
            title,
            oldview=showinitial,
            )

        # Bind only to newview.score because it alone takes focus.
        self.bind_buttons_to_widget(newview.score)

        self.ui = ui
        
    def dialog_ok(self):
        """Update record and return update action response (True for updated).

        Check that database is open and is same one as update action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text='Cannot update because not connected to a database')
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        self.newobject.value.load(
            repr(self.newview.get_name_query_statement_text()))
        if not len(self.newobject.value.get_name_text()):
            dialogues.showerror(
                title=self.__title,
                message=''.join((
                    "The selection rule has no name.\n\nPlease enter it's ",
                    "name as the first line of text.'")))
            return False
        if self.newobject.value.where_error:
            if tkinter.messagebox.YES != dialogues.askquestion(
                title=self.__title,
                message=''.join((
                    'Confirm request to update game selection rule named:\n\n',
                    self.newobject.value.get_name_text(),
                    '\n\non database.\n\n',
                    self.newobject.value.where_error.get_error_report(
                        self.ui.base_games.get_data_source())))):
                return False
        return super(ChessDBeditQuery, self).dialog_ok()

    def tidy_on_destroy(self):
        """Clear up after dialogue destruction."""
        self.ui.selections_in_toplevels.discard(self.oldview)
        self.ui.selections_in_toplevels.discard(self.newview)
        self.ui.base_selections.selection.clear()



# querydbdelete.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise delete dialogue to delete game selection rule record.
"""

from solentware_grid.gui.datadelete import DataDelete

from .chessexception import ChessException
from .querydisplay import DialogueQueryDisplay


class ChessDBdeleteQuery(ChessException, DataDelete):
    """Dialog to delete a game selection rule from database.

    The game selection rule is in it's own Toplevel widget.

    """

    def __init__(self, parent, instance, ui=None):
        """Extend and create dialogue widget for deleting selection rule."""
        oldview = DialogueQueryDisplay(master=parent, ui=ui)
        if ui is not None:
            ui.selections_in_toplevels.add(oldview)
            oldview.query_statement.set_database(
                ui.base_games.datasource.dbhome)
            oldview.query_statement.dbset = ui.base_games.datasource.dbset
        oldview.query_statement.process_query_statement(instance.get_srvalue())
        oldview.set_query_statement(instance.value)
        super(ChessDBdeleteQuery, self).__init__(
            instance,
            parent,
            oldview,
            ':  '.join((
                'Delete Selection Rule',
                instance.value._description_string)),
            )
        self.bind_buttons_to_widget(oldview.score)
        self.ui = ui
       
    def dialog_ok(self):
        """Delete record and return delete action response (True for deleted).

        Check that database is open and is same one as deletion action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text='Cannot delete because not connected to a database')
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        return super(ChessDBdeleteQuery, self).dialog_ok()

    def tidy_on_destroy(self):
        """Clear up after dialogue destruction."""
        self.ui.selections_in_toplevels.discard(self.oldview)
        self.ui.base_selections.selection.clear()



# gamedbdelete.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise delete dialogue to delete chess game record.
"""

from solentware_grid.gui.datadelete import DataDelete

from pgn_read.core.constants import TAG_WHITE, TAG_BLACK

from .chessexception import ChessException
from .gamedisplay import DialogueGameDisplay


class ChessDBdeleteGame(ChessException, DataDelete):
    
    """Dialog to delete a game from database.

    The game is in it's own Toplevel widget and playing through the game does
    not change the list of games, in the main widget, matching the current
    position on the board.

    """

    def __init__(self, parent, oldobject, ui=None):
        """Extend and create dialogue widget for deleting chess game."""
        oldview = DialogueGameDisplay(master=parent, ui=ui)
        oldview.set_position_analysis_data_source()
        if ui is not None:
            ui.games_and_repertoires_in_toplevels.add(oldview)
        oldview.pgn.get_first_game(oldobject.get_srvalue())
        oldview.set_game()
        tags = oldobject.value.collected_game[1]
        try:
            tt = '  '.join((
                'Delete Game:',
                ' - '.join((
                    tags[TAG_WHITE],
                    tags[TAG_BLACK])),
                ))
        except TypeError:
            tt = 'Delete Game - names unknown or invalid'
        except KeyError:
            tt = 'Delete Game - names unknown or invalid'
        super(ChessDBdeleteGame, self).__init__(oldobject, parent, oldview, tt)
        self.bind_buttons_to_widget(oldview.score)
        self.bind_buttons_to_widget(oldview.analysis.score)
        self.ui = ui

    def dialog_ok(self):
        """Delete record and return delete action response (True for deleted).

        Check that database is open and is same one as deletion action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text='Cannot delete because not connected to a database')
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        return super(ChessDBdeleteGame, self).dialog_ok()

    def delete(self):
        """Mark partial position records for recalculation and return key"""
        self.datasource.dbhome.mark_partial_positions_to_be_recalculated()
        super(ChessDBdeleteGame, self).delete()

    def tidy_on_destroy(self):
        """Clear up after dialogue destruction."""
        self.ui.games_and_repertoires_in_toplevels.discard(self.oldview)
        self.ui.base_games.selection.clear()



# uci_to_pgn.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Convert UCI Chess Engine moves to PGN.

UCI chess engines give moves as <from square><to square> and the easiest way of
handling these, given the existing PGN parser, is translate to PGN.

generate_pgn_for_uci_moves_in_position() generates the legal unambiguous PGN
description of a sequence of UCI Chess Engine move for the position it refers
to.

'Qd1f3' is always unambiguous but is the legal unambiguous description only if
more than two queens of the side with the move can legally move to 'f3'. 'Qf3'
is usually the only legal PGN description because the starting position for a
game has one queen per side. 'Qf3' is not necessarily a legal move.

"""
import re

from pgn_read.core.parser import PGNMove
from pgn_read.core.constants import (
    WKING,
    WQUEEN,
    WROOK,
    WBISHOP,
    WKNIGHT,
    WPAWN,
    BKING,
    BQUEEN,
    BROOK,
    BBISHOP,
    BKNIGHT,
    BPAWN,
    PGN_KING,
    PGN_QUEEN,
    PGN_ROOK,
    PGN_BISHOP,
    PGN_KNIGHT,
    PGN_PAWN,
    SPACE,
    NOPIECE,
    MAP_PGN_SQUARE_NAME_TO_FEN_ORDER,
    CAPTURE_MOVE,
    PLAIN_MOVE,
    PIECE_CAPTURE_MAP,
    PIECE_MOVE_MAP,
    SQUARE_BITS,
    SIDE_KING,
    OTHER_SIDE,
    GAPS,
    )

_PIECE_TO_PGN = {
    WKING: PGN_KING,
    WQUEEN: PGN_QUEEN,
    WROOK: PGN_ROOK,
    WBISHOP: PGN_BISHOP,
    WKNIGHT: PGN_KNIGHT,
    WPAWN: PGN_PAWN,
    BKING: PGN_KING,
    BQUEEN: PGN_QUEEN,
    BROOK: PGN_ROOK,
    BBISHOP: PGN_BISHOP,
    BKNIGHT: PGN_KNIGHT,
    BPAWN: PGN_PAWN,
    }
_PROMOTE = {
    PGN_QUEEN.lower(): ''.join(('=', PGN_QUEEN)),
    PGN_ROOK.lower(): ''.join(('=', PGN_ROOK)),
    PGN_BISHOP.lower(): ''.join(('=', PGN_BISHOP)),
    PGN_KNIGHT.lower(): ''.join(('=', PGN_KNIGHT)),
    '': '',
    }
_PROMOTE_PIECE = {
    ''.join(('8', PGN_QUEEN.lower())): WQUEEN,
    ''.join(('8', PGN_ROOK.lower())): WROOK,
    ''.join(('8', PGN_BISHOP.lower())): WBISHOP,
    ''.join(('8', PGN_KNIGHT.lower())): WKNIGHT,
    ''.join(('1', PGN_QUEEN.lower())): BQUEEN,
    ''.join(('1', PGN_ROOK.lower())): BROOK,
    ''.join(('1', PGN_BISHOP.lower())): BBISHOP,
    ''.join(('1', PGN_KNIGHT.lower())): BKNIGHT,
    }
_CASTLES = {'e1g1': 'O-O', 'e8g8': 'O-O', 'e1c1': 'O-O-O', 'e8c8': 'O-O-O'}
_CASTLEKEY = {
    'e1g1': WKING,
    'e8g8': BKING,
    'e1c1': WKING,
    'e8c8': BKING}
_FEN_SPACES = {
    '1': SPACE,
    '2': SPACE * 2,
    '3': SPACE * 3,
    '4': SPACE * 4,
    '5': SPACE * 5,
    '6': SPACE * 6,
    '7': SPACE * 7,
    '8': SPACE * 8,
    }

re_move = re.compile(''.join(('^',
                              '([a-h][1-8])',
                              '([a-h][1-8])',
                              '([qrbn]?)',
                              '$')))


def generate_pgn_for_uci_moves_in_position(moves, fen):
    """Return PGN-style movetext and update position for unambiguous moves.

    """
    pgn = PGNMove()
    text = []
    try:
        moves = moves.split()
    except:
        return ''.join(
            ("{'",
             str(moves),
             "' cannot be a move, 'Yz0' inserted.}Yz0"))
    if not moves:
        return "{'' is not a move, 'Yz0' inserted. Rest '' ignored.}Yz0"
    pgn.set_position_fen(fen)
    if not pgn._initial_fen:
        return ''.join((
            "{'Forsyth-Edwards Notation sets an illegal position. ",
            "Move 'Yz0' inserted.}Yz0"))
    for count, move in enumerate(moves):
        g = re_move.match(move)
        if not g:
            text.append(''.join(
                ("{'",
                 str(move),
                 "' cannot be a move, 'Yz0' inserted. Rest '",
                 ' '.join(moves[count+1:]),
                 "' ignored.}Yz0")))
            break
        from_square, to_square, promote_to = g.groups()
        fen_from_square = MAP_PGN_SQUARE_NAME_TO_FEN_ORDER[from_square]
        fen_to_square = MAP_PGN_SQUARE_NAME_TO_FEN_ORDER[to_square]
        gap = GAPS[fen_from_square][fen_to_square]

        # from_square must contain a piece belonging to side with the move.
        if fen_from_square not in pgn.occupied_squares[pgn.active_side]:
            text.append(''.join(
                ("{'",
                 str(move),
                 "' does not refer to a piece of the active side, ",
                 "'Yz0' inserted. Rest '",
                 ' '.join(moves[count+1:]),
                 "' ignored.}Yz0")))
            break

        piece = _PIECE_TO_PGN[pgn.board[fen_from_square]]
        promote_to = _PROMOTE[promote_to]

        # What about illegal moves specified in e2f5 format where the
        # constructed PGN move happens to be legal.

        # Castles is treated as a king move of two squares, otherwise illegal.
        if (move in _CASTLES and
            pgn.board[fen_from_square] == _CASTLEKEY.get(move)):
            pgn_move_descriptions = _CASTLES[move],

        # Pawns have moving rules different from the other pieces.
        elif piece == PGN_PAWN:

            # The distinction petween normal captures, to_square occupied, and
            # en-passant captures, to_square empty, is left to the PGN parser.
            if from_square[0] == to_square[0]:
                takes = PLAIN_MOVE
                move_map = PIECE_MOVE_MAP[pgn.board[fen_from_square]]
            else:
                takes = CAPTURE_MOVE
                move_map = PIECE_CAPTURE_MAP[pgn.board[fen_from_square]]

            # Pawn must be able to make the move: nothing in the way.
            if gap & pgn.board_bitmap or (not move_map[fen_to_square] &
                                          SQUARE_BITS[fen_from_square]):
                text.append(''.join(
                    ("{'",
                     str(move),
                     "' cannot be a move, 'Yz0' inserted. Rest '",
                     ' '.join(moves[count+1:]),
                     "' ignored.}Yz0")))
                break

            # Pawn must not be promoted to ranks othen than 1 or 8.
            if promote_to:
                if to_square[1] not in '18':
                    text.append(''.join(
                        ("{'",
                         str(move),
                         "' cannot be a move, 'Yz0' inserted. Rest '",
                         ' '.join(moves[count+1:]),
                         "' ignored.}Yz0")))
                    break

            # A pawn move description is unambigous so validation that 'move'
            # does not leave king in check is left to PGN parser.

            if takes == PLAIN_MOVE:
                pgn_move_descriptions = ''.join((to_square, promote_to)),
            else:
                pgn_move_descriptions = (
                    ''.join((from_square[0], takes, to_square, promote_to)),
                    )
        
        else:
            if pgn.board[fen_to_square] != NOPIECE:
                takes = CAPTURE_MOVE
                move_map = PIECE_CAPTURE_MAP[pgn.board[fen_from_square]]
            else:
                takes = PLAIN_MOVE
                move_map = PIECE_MOVE_MAP[pgn.board[fen_from_square]]

            # Piece must be able to make the move: nothing in the way.
            if gap & pgn.board_bitmap or (not move_map[fen_to_square] &
                                          SQUARE_BITS[fen_from_square]):
                text.append(''.join(
                    ("{'",
                     str(move),
                     "' cannot be a move, 'Yz0' inserted. Rest '",
                     ' '.join(moves[count+1:]),
                     "' ignored.}Yz0")))
                break

            # If more than one piece of a type is able to move to a square but
            # some of them are pinned against their king, the PGN description
            # of the move must not include the elements of from_square usually
            # given to distinguish the moves.
            # The piece on from_square must not be pinned against it's king
            # unless it's move does not break the pin.
            if len(pgn.piece_locations[pgn.board[fen_from_square]]) > 1:
                for active_king_square in pgn.piece_locations[
                    SIDE_KING[pgn.active_side]]:
                    pass # bind active_king_square without pop() and add()
                inactive_pieces = pgn.occupied_squares[
                    OTHER_SIDE[pgn.active_side]].copy()
                inactive_pieces.discard(fen_to_square)
                bitmap = pgn.board_bitmap & (pgn.board_bitmap ^
                                             SQUARE_BITS[fen_from_square])
                bitmap |= SQUARE_BITS[fen_to_square]
                gk = GAPS[active_king_square]
                for square in inactive_pieces:
                    if (not gk[square] & bitmap and
                        PIECE_CAPTURE_MAP[pgn.board[square]
                                          ][active_king_square] &
                        SQUARE_BITS[square]):
                        text.append(''.join(
                            ("{'",
                             str(move),
                             "' cannot be a move, 'Yz0' inserted. Rest '",
                             ' '.join(moves[count+1:]),
                             "' ignored.}Yz0")))
                        pinned = True
                        break
                else:
                    pinned = False
                if pinned:
                    break

            pgn_move_descriptions = (
                ''.join((piece, takes, to_square, promote_to)),
                ''.join((piece, from_square[0], takes, to_square, promote_to)),
                ''.join((piece, from_square[1], takes, to_square, promote_to)),
                ''.join((piece, from_square, takes, to_square, promote_to)),
                )

        # Use the first PGN move description accepted by PGN parser.
        for pgn_move in pgn_move_descriptions:
            ravstack = pgn.ravstack[-1]
            t = pgn.get_first_pgn_token(pgn_move)
            if not t:
                text.append(''.join(
                    ("{'",
                     str(move),
                     "' cannot be a move, 'Yz0' inserted. Rest '",
                     ' '.join(moves[count+1:]),
                     "' ignored.}Yz0")))
                break
            if not pgn.error_tokens:
                text.append(pgn_move)
                break
            pgn.error_tokens = []
            pgn.ravstack[-1] = ravstack
            pgn.reset_position(ravstack[-1])
        else:
            text.append(''.join(
                ("{'",
                 str(move),
                 "' cannot be a move, 'Yz0' inserted. Rest '",
                 ' '.join(moves[count+1:]),
                 "' ignored.}Yz0")))
            break

    return ' '.join(text)



# cqldbedit.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Edit or insert dialogue for Chess Query Language (ChessQL) statement record.
"""
import tkinter

from solentware_misc.workarounds import dialogues

from solentware_grid.gui.dataedit import DataEdit

from .chessexception import ChessException
from .cqldisplay import DialogueCQLDisplay, DialogueCQLEdit


class ChessDBeditCQL(ChessException, DataEdit):
    """Dialog to edit a ChessQL statement on, or insert one into, database.

    The ChessQL statement is in it's own Toplevel widget.

    """

    def __init__(self, newobject, parent, oldobject, showinitial=True, ui=None):
        """Extend and create dialogue to edit or insert ChessQL statement."""
        if oldobject:
            title = ':  '.join((
                'Edit ChessQL statement',
                oldobject.value.get_name_text()))
        else:
            title = 'Insert ChessQL Statement'
            showinitial = False
        self.__title = title.split(':')[0]
        if showinitial:
            showinitial = DialogueCQLDisplay(master=parent, ui=ui)
            if ui is not None:
                ui.partials_in_toplevels.add(showinitial)
            showinitial.cql_statement.process_statement(
                oldobject.get_srvalue())
            showinitial.set_cql_statement()
        newview = DialogueCQLEdit(master=parent, ui=ui)
        if ui is not None:
            ui.partials_in_toplevels.add(newview)
        newview.cql_statement.process_statement(newobject.get_srvalue())
        newview.set_cql_statement()
        super().__init__(
            newobject,
            parent,
            oldobject,
            newview,
            title,
            oldview=showinitial,
            )

        # Bind only to newview.score because it alone takes focus.
        self.bind_buttons_to_widget(newview.score)

        self.ui = ui
        
    def dialog_ok(self):
        """Update record and return update action response (True for updated).

        Check that database is open and is same one as update action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text='Cannot update because not connected to a database')
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        self.newobject.value.load(
            repr(self.newview.get_name_cql_statement_text()))
        if not len(self.newobject.value.get_name_text()):
            if not self.newobject.value.cql_error:
                dialogues.showerror(
                    title=self.__title,
                    message=''.join((
                        "The ChessQL statement has no name.\n\nPlease enter ",
                        "it's name as the first line of text.'")))
            else:
                dialogues.showerror(
                    title=self.__title,
                    message=''.join((
                        "The text does not contain a valid ChessQL ",
                        "statement. ")))
            return False
        elif self.newobject.value.cql_error:
            if tkinter.messagebox.YES != dialogues.askquestion(
                title=self.__title,
                message=''.join((
                    'Confirm request to update ChessQL statement named:\n\n',
                    self.newobject.value.get_name_text(),
                    '\n\non database.\n\n',
                    self.newobject.value.cql_error.get_error_report()))):
                return False
        if self.ui.partial_items.active_item:
            if self.ui.partial_items.active_item.sourceobject is None:
                dialogues.showinfo(
                    title=self.__title,
                    message=''.join((
                        "Cannot use this insert dialogue while the active ",
                        "item in cql queries is one opened by menu action ",
                        "'Position | Partial'.")))
                return False
        return super().dialog_ok()

    def tidy_on_destroy(self):
        """Clear up after dialogue destruction."""
        self.ui.partials_in_toplevels.discard(self.oldview)
        self.ui.partials_in_toplevels.discard(self.newview)
        self.ui.base_partials.selection.clear()

    def edit(self, commit=True):
        """Edit the records and refresh widgets.

        The ChessQL query statement update is delegated to the superclass.

        This method updates the foundset calculated from the query.

        """
        if commit:
            self.datasource.dbhome.start_transaction()
        super().edit(commit=False)
        cqls = self.ui.partialpositionds(
            self.ui.base_games.datasource.dbhome,
            self.ui.base_games.datasource.dbset,
            self.ui.base_games.datasource.dbset,
            newrow=None)
        cqls.update_cql_statement_games(self.newobject, commit=False)
        if commit:
            self.datasource.dbhome.commit()
        
    def put(self, commit=True):
        """Insert the records and refresh widgets.

        The ChessQL query statement update is delegated to the superclass.

        This method updates the foundset calculated from the query.

        """
        if commit:
            self.datasource.dbhome.start_transaction()
        super().put(commit=False)
        cqls = self.ui.partialpositionds(
            self.ui.base_games.datasource.dbhome,
            self.ui.base_games.datasource.dbset,
            self.ui.base_games.datasource.dbset,
            newrow=None)
        cqls.update_cql_statement_games(self.newobject, commit=False)
        if commit:
            self.datasource.dbhome.commit()



# cqldbdelete.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Delete dialogue for Chess Query Language (ChessQL) statement record.
"""

from solentware_misc.workarounds import dialogues

from solentware_grid.gui.datadelete import DataDelete

from .chessexception import ChessException
from .cqldisplay import DialogueCQLDisplay


class ChessDBdeleteCQL(ChessException, DataDelete):
    """Dialog to delete a ChessQL statement from database.

    The ChessQL statement is in it's own Toplevel widget.

    """

    def __init__(self, parent, instance, ui=None):
        """Extend and create dialogue widget for deleting ChessQL statement."""
        title = ':  '.join((
            'Delete ChessQL statement',
            instance.value.get_name_text()))
        self.__title = title.split(':')[0]
        oldview = DialogueCQLDisplay(master=parent, ui=ui)
        if ui is not None:
            ui.partials_in_toplevels.add(oldview)
        oldview.cql_statement.process_statement(instance.get_srvalue())
        oldview.set_cql_statement(instance.value)
        super().__init__(
            instance,
            parent,
            oldview,
            title,
            )
        self.bind_buttons_to_widget(oldview.score)
        self.ui = ui
       
    def dialog_ok(self):
        """Delete record and return delete action response (True for deleted).

        Check that database is open and is same one as deletion action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text='Cannot delete because not connected to a database')
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        if self.ui.partial_items.active_item:
            if self.ui.partial_items.active_item.sourceobject is None:
                dialogues.showinfo(
                    title=self.__title,
                    message=''.join((
                        "Cannot use this delete dialogue while the active ",
                        "item in cql queries is one opened by menu action ",
                        "'Position | Partial'.")))
                return False
            if (self.ui.partial_items.active_item.sourceobject.key ==
                self.object.key):
                dialogues.showinfo(
                    title=self.__title,
                    message=''.join((
                        "Cannot use this delete dialogue while the active ",
                        "item in cql queries is one that displays the record ",
                        "being deleted.")))
                return False
        return super().dialog_ok()

    def tidy_on_destroy(self):
        """Clear up after dialogue destruction."""
        self.ui.partials_in_toplevels.discard(self.oldview)
        self.ui.base_partials.selection.clear()

    def delete(self, commit=True):
        """Delete the record and refresh widgets.

        The ChessQL query statement update is delegated to the superclass.

        This method deletes the foundset calculated from the query.

        """
        if commit:
            self.datasource.dbhome.start_transaction()
        super().delete(commit=False)
        cqls = self.ui.partialpositionds(
            self.ui.base_games.datasource.dbhome,
            self.ui.base_games.datasource.dbset,
            self.ui.base_games.datasource.dbset,
            newrow=None)
        assert self.object.newrecord is None
        cqls.forget_cql_statement_games(self.object, commit=False)
        if commit:
            self.datasource.dbhome.commit()



# repertoiredbedit.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise edit dialogue to edit or insert repertoire record.
"""

import tkinter

from solentware_grid.gui.dataedit import DataEdit

from solentware_misc.workarounds import dialogues

from pgn_read.core.constants import TAG_OPENING

from .chessexception import ChessException
from .repertoiredisplay import DialogueRepertoireDisplay, DialogueRepertoireEdit
from .constants import EMPTY_REPERTOIRE_GAME


class ChessDBeditRepertoire(ChessException, DataEdit):
    
    """Dialog to edit a repertoire on, or insert a repertoire into, database.

    The repertoire is in it's own Toplevel widget and playing through it does
    not change the list of games, in the main widget, matching the current
    position on the board.

    """

    def __init__(self, newobject, parent, oldobject, showinitial=True, ui=None):
        """Extend and create dialogue widget to edit or insert chess game."""
        if oldobject:
            try:
                title = '  '.join((
                    'Edit Repertoire:',
                    oldobject.value.collected_game[1][TAG_OPENING],
                    ))
            except TypeError:
                title = 'Edit Repertoire - name unknown or invalid'
            except KeyError:
                title = 'Edit Repertoire - name unknown or invalid'
        else:
            title = 'Insert Repertoire'
            showinitial = False
        if showinitial:
            showinitial = DialogueRepertoireDisplay(master=parent, ui=ui)
            showinitial.set_position_analysis_data_source()
            if ui is not None:
                ui.games_and_repertoires_in_toplevels.add(showinitial)
            showinitial.pgn.get_first_game(newobject.get_srvalue())
            showinitial.set_game()
        newview = DialogueRepertoireEdit(master=parent, ui=ui)
        newview.set_position_analysis_data_source()
        if ui is not None:
            ui.games_and_repertoires_in_toplevels.add(newview)
        if oldobject:
            newview.pgn.get_first_game(newobject.get_srvalue())
            oldobject.value.set_game_source('No opening name')
        else:
            newview.pgn.get_first_game(
                ''.join((EMPTY_REPERTOIRE_GAME, '*')))
        newview.set_game()
        super(ChessDBeditRepertoire, self).__init__(
            newobject,
            parent,
            oldobject,
            newview,
            title,
            oldview=showinitial,
            )

        # Bind only to newview.score and newview.analysis.score because these
        # alone takes focus.
        self.bind_buttons_to_widget(newview.score)
        self.bind_buttons_to_widget(newview.analysis.score)

        self.ui = ui

    def dialog_ok(self):
        """Update record and return update action response (True for updated).

        Check that database is open and is same one as update action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text='Cannot update because not connected to a database')
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        text = self.newview.get_score_error_escapes_removed()
        self.newobject.value.load(repr(text))
        if not self.newobject.value.is_pgn_valid():
            print('epertoiredbedit dialog_ok')
            if tkinter.messagebox.YES != dialogues.askquestion(
                title='Edit Repertoire',
                message=''.join(
                    ('The edited repertoire contains at least one illegal ',
                     'move in PGN.\n\nPlease re-confirm request to edit ',
                     'repertoire.',
                     ))):
                return False
            self.newobject.value.set_game_source('No opening name')
        return super(ChessDBeditRepertoire, self).dialog_ok()

    def tidy_on_destroy(self):
        """Clear up after dialogue destruction."""
        self.ui.games_and_repertoires_in_toplevels.discard(self.oldview)
        self.ui.games_and_repertoires_in_toplevels.discard(self.newview)
        self.ui.base_repertoires.selection.clear()



# analysis.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess engine analysis for a position.
"""

import sys

from pgn_read.core.constants import (
    TAG_FEN,
    UNKNOWN_RESULT,
    END_TAG,
    START_TAG,
    START_RAV,
    END_RAV,
    START_COMMENT,
    END_COMMENT,
    )


class Analysis(object):
    """Chess engine analysis for FEN position.

    Store variations generated by chess engines for a position.

    """

    def __init__(self, position=None):
        """"""
        super().__init__()

        # Position encoded as index value for position index.
        self.position = position

        # Variations generated by chess engines:
        # key is engine name reported by chess engine when started,
        # value is ((evaluation, PGN of variation), ...) sorted by evaluation.
        # len(value) is multiPV option value used to do analysis.
        self.variations = {}

        # Depth and width of analysis by chess engine
        # key is engine name reported by chess engine when started,
        # value is (depth, multiPV) option values used to do analysis.
        self.scale = {}

    def set_variations_empty(self):
        """Set all variations to None.

        len(variations[engine] is the multiPV value used when analysis done.
        Set variations to None to preserve the multiPV information when a new
        analysis request is done, while noting any analysis deemed necessary
        is not yet done.

        """
        var = self.variations
        for k, v in var.items():
            var[k] = [None] * len(v)

    def translate_analysis_to_pgn(self, move_played=''):
        """Translate UCI chess engine variation output to PGN."""

        # Addition of this code tips balance in favour of a class to deal with
        # FEN representation of positions.
        to_move = self.position.split()[1]
        
        if move_played:
            move_played = ''.join(
                (move_played,
                 ' ;Move played',
                 ' by white' if to_move == 'w' else
                 ' by black' if to_move == 'b' else
                 '',
                 '\n'))
        variations = self.variations
        scale = self.scale
        new_text = []
        for engine_name in sorted(variations):

            # The following stopped happening after the test database was
            # deleted and re-created, so maybe everything is sound and the
            # problem is left-over from earlier testing.
            #
            # This may happen if communication with a remote engine is lost
            # and then re-established by 'quit engine' and 'start engine' menu
            # actions, or perhaps if the process driving a local engine fails
            # too.  It has been seen only for the position on display in the
            # active window at the time these actions are done while the
            # database is open.
            # The sequence of actions is:
            # Start a remote UCI server and open a database in either order.
            # Use the Engine menu to start a UCI client for the UCI server.
            # Close the database (so no analysis get written to database).
            # Use the Game menu to start a new game and see analysis appear.
            # Stop the remote UCI server (action on the remote host).
            # Start a remote UCI server.
            # Open a database.
            # Use the Engine menu to start a UCI client for the UCI server.
            # Now closing and opening the database switches it off and on.
            #if engine_name is None:
            #    sys.stderr.write(
            #        'Name of engine: ' + str(engine_name) + ' ' + str(len(variations)) + '\n')
            #    continue

            analysis = variations[engine_name]

            # Keep going as normal for non-str engine_name (should not happen).
            #new_text.append(engine_name.join((';', '\n')))
            new_text.append(str(engine_name).join((';', '\n')))

            depth, multipv = [str(s) for s in scale[engine_name]]
            lines = [''.join((' '.join((START_COMMENT,
                                        self._evalution_score(a[0], to_move),
                                        depth,
                                        str(e + 1),
                                        ' '.join((END_COMMENT, START_RAV)))),
                              a[1],
                              END_RAV,
                              '\n'))
                     for e, a in enumerate(analysis)]
            new_text.extend(lines)
            if not move_played:
                move_played = ''.join(
                    (analysis[0][1].split()[0],
                     ' ;First variation',
                     ', white to move' if to_move == 'w' else
                     ', black to move' if to_move == 'b' else
                     '',
                     '\n'))
        new_text.append(UNKNOWN_RESULT)
        if move_played:
            new_text.insert(0, move_played)
            new_text.insert(0,
                            ''.join(
                                (START_TAG, TAG_FEN, '"',
                                 self.position,
                                 END_TAG.join('"\n'))))
        return ''.join(new_text)

    def _evalution_score(self, val, to_move):
        """Normalize to white advantage is positive and black negative."""
        try:
            return '{:+.2f}'.format(
                (int(val) if to_move != 'b' else -int(val)) / 100)
        except:
            return '?.??'



# setup.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

from setuptools import setup

if __name__ == '__main__':

    long_description = open('README').read()

    setup(
        name='chesstab',
        version='4.2',
        description='Database for chess games',
        author='Roger Marsh',
        author_email='roger.marsh@solentware.co.uk',
        url='http://www.solentware.co.uk',
        packages=[
            'chesstab',
            'chesstab.core', 'chesstab.basecore', 'chesstab.gui',
            'chesstab.help',
            'chesstab.db', 'chesstab.dpt', 'chesstab.sqlite', 'chesstab.apsw',
            'chesstab.unqlite', 'chesstab.vedis',
            'chesstab.gnu', 'chesstab.ndbm',
            'chesstab.fonts',
            'chesstab.tools',
            ],
        package_data={
            'chesstab.fonts': ['*.TTF', '*.zip'],
            'chesstab.help': ['*.rst', '*.html'],
            },
        long_description=long_description,
        license='BSD',
        classifiers=[
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Operating System :: OS Independent',
            'Topic :: Games/Entertainment :: Board Games',
            'Intended Audience :: End Users/Desktop',
            'Development Status :: 3 - Alpha',
            ],
        install_requires=[
            'solentware-base==4.1.2',
            'chessql==2.0',
            'solentware-grid==2.0',
            'pgn-read==1.3.2',
            'solentware-misc==1.2',
            'uci-net==1.2',
            ],
        dependency_links=[
            'http://solentware.co.uk/files/solentware-base-4.1.2.tar.gz',
            'http://solentware.co.uk/files/chessql-2.0.tar.gz',
            'http://solentware.co.uk/files/solentware-grid-2.0.tar.gz',
            'http://solentware.co.uk/files/pgn-read-1.3.2.tar.gz',
            'http://solentware.co.uk/files/solentware-misc-1.2.tar.gz',
            'http://solentware.co.uk/files/uci-net-1.2.tar.gz',
            ],
        )



# colourscheme.py
# Copyright 2009 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""A chess board colour scheme chooser.
"""

import tkinter
import tkinter.font

from solentware_misc.gui.colourslider import ColourSlider

from pgn_read.core.parser import PGNDisplay

from .chessexception import ChessException
from . import constants, fonts, options
from . import game, board


class _ColourScheme(ChessException):

    """Widgets to select fonts and colours for chess board and game score.

    Use the subclasses defined in this module depending on what fits on
    the screen.
    
    """

    def __init__(
        self,
        ui=None,
        height=500,
        title='Style',
        **kargs):
        """Create Toplevel and fonts to be modified.
        
        ui is the ChessUI instance
        height - passed to Tkinter.Toplevel.
        title - use is w.wm_title(title).

        """
        self.ui = ui
        tags = ui.tags_variations_comments_font
        moves = ui.moves_played_in_game_font
        pieces = ui.boardfont
        self.ok = False
        if tags:
            self.tags_variations_comments_font = tags.copy()
        else:
            self.tags_variations_comments_font = tkinter.font.nametofont(
                constants.TAGS_VARIATIONS_COMMENTS_FONT).copy()
        if moves:
            self.moves_played_in_game_font = moves.copy()
        else:
            self.moves_played_in_game_font = tkinter.font.nametofont(
                constants.MOVES_PLAYED_IN_GAME_FONT).copy()
        if pieces:
            self.boardfont = pieces.copy()
        else:
            self.boardfont = tkinter.font.nametofont(
                constants.PIECES_ON_BOARD_FONT).copy()
        # Wildpieces font is based on moves font but managed like boardfont
        if moves:
            self.wildpiecesfont = moves.copy()
        else:
            self.wildpiecesfont = tkinter.font.nametofont(
                constants.WILDPIECES_ON_BOARD_FONT).copy()
        self.chooser = tkinter.Toplevel(
            height=height,
            width=700)
        self.chooser.wm_resizable(False, False)
        self.chooser.wm_title(title)
        self.chooser.pack_propagate(False)

    def apply_modified_font(self, modifiedfont, font):
        """Copy modifiedfont attributes to font."""
        if self.is_ok():
            fonts.copy_font_attributes(modifiedfont.actual(), font)

    def apply_modified_board_font(self, modifiedfont, font):
        """Copy modifiedfont attributes to board font."""
        if self.is_ok():
            fonts.copy_board_font_attributes(modifiedfont.actual(), font)

    def apply_pieces_on_board_font_to_font(self, font):
        """Copy self.boardfont attributes to board font."""
        self.apply_modified_board_font(self.boardfont, font)

    def apply_to_named_fonts(self):
        """Apply modified fonts to all displayed games."""
        self.apply_to_named_font_moves_played_in_game()
        self.apply_to_named_font_pieces_on_board()
        self.apply_to_named_font_wildpieces_on_board()
        self.apply_to_named_font_tags_variations_comments()

    def apply_to_named_font_moves_played_in_game(self):
        """Apply moves played font to all displayed games."""
        self.apply_modified_font(
            self.moves_played_in_game_font,
            tkinter.font.nametofont(constants.MOVES_PLAYED_IN_GAME_FONT)
            )

    def apply_to_named_font_pieces_on_board(self):
        """Apply pieces font to all displayed games."""
        self.apply_modified_board_font(
            self.boardfont,
            tkinter.font.nametofont(constants.PIECES_ON_BOARD_FONT)
            )

    def apply_to_named_font_tags_variations_comments(self):
        """Apply variations and comments font to all displayed games."""
        self.apply_modified_font(
            self.tags_variations_comments_font,
            tkinter.font.nametofont(constants.TAGS_VARIATIONS_COMMENTS_FONT)
            )

    def apply_to_named_font_wildpieces_on_board(self):
        """Apply wildpieces font to all displayed games."""
        self.apply_modified_board_font(
            self.wildpiecesfont,
            tkinter.font.nametofont(constants.WILDPIECES_ON_BOARD_FONT)
            )

    def get_button_definitions(self):
        """Return button definitions."""
        return (
            ('OK',
             'OK button Tooltip.',
             True,
             -1,
             self.on_ok),
            ('Cancel',
             'Cancel button Tooltip.',
             True,
             2,
             self.on_cancel),
            )

    def create_buttons(self, buttons):
        """Create buttons."""
        self.buttons_frame = tkinter.Frame(master=self.chooser)
        self.buttons_frame.pack(side=tkinter.BOTTOM, fill=tkinter.X)
        buttonrow = self.buttons_frame.pack_info()['side'] in ('top', 'bottom')
        for i, b in enumerate(buttons):
            button = tkinter.Button(
                master=self.buttons_frame,
                text=buttons[i][0],
                underline=buttons[i][3],
                command=self.try_command(buttons[i][4], self.buttons_frame))
            if buttonrow:
                self.buttons_frame.grid_columnconfigure(i*2, weight=1)
                button.grid_configure(column=i*2 + 1, row=0)
            else:
                self.buttons_frame.grid_rowconfigure(i*2, weight=1)
                button.grid_configure(row=i*2 + 1, column=0)
        if buttonrow:
            self.buttons_frame.grid_columnconfigure(
                len(buttons*2), weight=1)
        else:
            self.buttons_frame.grid_rowconfigure(
                len(buttons*2), weight=1)

    def create_colour_frame(self):
        """Create colour selector widgets."""
        colour_frame = tkinter.Frame(master=self.chooser)
        colour_frame.pack(fill=tkinter.X)
        gs = self.game.score
        gb = self.game.board
        for row, label, color, setter in (
            (0, 'Light squares', gb.litecolor, self.set_lite_square_bgcolour),
            (1, 'Dark squares', gb.darkcolor, self.set_dark_square_bgcolour),
            (2, 'White pieces', gb.whitecolor, self.set_white_piece_fgcolour),
            (3, 'Black pieces', gb.blackcolor, self.set_black_piece_fgcolour),
            (4, 'Rest of line', self.game.l_color, self.set_tag_rest_of_line),
            (5, 'Move', self.game.m_color, self.set_tag_move),
            (6, 'Alternatives', self.game.am_color, self.set_tag_alternatives),
            (7, 'Start of line', self.game.v_color, self.set_tag_start_of_line),
            ):
            ChessColourSlider(
                colour_frame,
                row=row,
                colour=color,
                label=label,
                TagConfigureBGcolour=setter)

    def create_font_frame(self):
        """Create font selection widget."""

        def focus(w):
            def fs(e=None):
                w.focus_set()
            return fs
        
        font_frame = tkinter.Frame(master=self.chooser)
        font_frame.pack(fill=tkinter.X)
        # font chooser for board
        font_frame.columnconfigure(
            0, weight=1, uniform='fontpanels')
        tkinter.Label(
            master=font_frame, text='Pieces').grid_configure(
                row=0, column=0)
        fontfr = tkinter.Frame(master=font_frame, padx=5)
        fr = tkinter.Frame(master=fontfr)
        self.b_families = tkinter.Listbox(fr)
        scrollfont = tkinter.Scrollbar(fr)
        scrollfont.configure(
            command=self.try_command(self.b_families.yview, scrollfont))
        self.b_families.configure(yscrollcommand=scrollfont.set)
        for f in sorted(tkinter.font.families()):
            self.b_families.insert(tkinter.END, f)
        self.b_families.pack(
            side=tkinter.LEFT, expand=tkinter.TRUE, fill=tkinter.Y)
        scrollfont.pack(side=tkinter.LEFT, fill=tkinter.Y)
        fr.pack(fill=tkinter.Y, expand=tkinter.TRUE)
        wssf = tkinter.Frame(master=fontfr)
        self.b_family = tkinter.Label(
            master=wssf, text=self.boardfont['family'])
        self.b_family.grid_configure(column=0, row=1, sticky=tkinter.EW)
        self.b_weight = tkinter.IntVar()
        cb = tkinter.Checkbutton(
            master=wssf,
            text='Bold',
            variable=self.b_weight,
            command=self.try_command(self.set_board_font_weight, wssf),
            indicatoron=tkinter.FALSE
            )
        cb.grid_configure(column=0, row=3, sticky=tkinter.EW)
        cb.bind('<ButtonPress>', self.try_event(focus(cb)))
        self.b_weight.set(tkinter.TRUE)
        wssf.grid_columnconfigure(0, weight=1)
        wssf.grid_rowconfigure(0, minsize=5)
        wssf.grid_rowconfigure(2, minsize=5)
        wssf.grid_rowconfigure(4, minsize=5)
        wssf.pack(fill=tkinter.X)
        fontfr.grid_configure(row=1, column=0, sticky=tkinter.NS)
        self.b_families.bind(
            '<<ListboxSelect>>', self.try_event(self.set_board_font_family))
        # font chooser for score
        font_frame.columnconfigure(
            1, weight=2, uniform='fontpanels')
        tkinter.Label(
            master=font_frame, text='Score').grid_configure(
                row=0, column=1)
        fontfr = tkinter.Frame(master=font_frame, padx=5)
        fr = tkinter.Frame(master=fontfr)
        self.s_families = tkinter.Listbox(fr)
        scrollfont = tkinter.Scrollbar(fr)
        scrollfont.configure(
            command=self.try_command(self.s_families.yview, scrollfont))
        self.s_families.configure(yscrollcommand=scrollfont.set)
        for f in sorted(tkinter.font.families()):
            self.s_families.insert(tkinter.END, f)
        self.s_families.pack(
            side=tkinter.LEFT, expand=tkinter.TRUE, fill=tkinter.X)
        scrollfont.pack(side=tkinter.LEFT, fill=tkinter.Y)
        fr.pack()
        wssf = tkinter.Frame(master=fontfr)
        self.s_family = tkinter.Label(
            master=wssf, text=self.tags_variations_comments_font['family'])
        self.s_family.grid_configure(
            column=1, row=1, sticky=tkinter.EW, columnspan=3)
        self.s_weight = tkinter.IntVar()
        self._make_font_selector(
            wssf, 'Bold:', self.s_weight, 3, self.set_score_font_weight)
        self.s_weight.set(self.get_combination('weight', tkinter.font.BOLD))
        self.s_slant = tkinter.IntVar()
        self._make_font_selector(
            wssf, 'Italic:', self.s_slant, 6, self.set_score_font_slant)
        self.s_slant.set(self.get_combination('slant', tkinter.font.ITALIC))
        wssf.grid_rowconfigure(0, minsize=5)
        wssf.grid_rowconfigure(2, minsize=5)
        wssf.grid_rowconfigure(5, minsize=5)
        wssf.grid_rowconfigure(8, minsize=5)
        wssf.grid_rowconfigure(10, minsize=5)
        wssf.grid_columnconfigure(1, weight=1, uniform='swght')
        wssf.grid_columnconfigure(2, weight=1, uniform='swght')
        wssf.grid_columnconfigure(3, weight=1, uniform='swght')
        self.s_size = tkinter.IntVar()
        tkinter.Label(
            master=wssf,
            text='Size:'
            ).grid_configure(column=0, row=9)
        sf = tkinter.Frame(master=wssf)
        fsize = self.tags_variations_comments_font.cget('size')
        sf.grid_configure(column=1, row=9, columnspan=3, sticky=tkinter.EW)
        for i, s in enumerate((7, 8, 9, 10, 11, 12, 13, 14, 16, 18, 20)):
            sb = tkinter.Radiobutton(
                master=sf,
                text=str(s),
                variable=self.s_size,
                value=s,
                indicatoron=tkinter.FALSE,
                command=self.try_command(self.set_score_font_size, sf))
            if s == abs(fsize):
                self.s_size.set(s)
            sb.grid_configure(column=i + 1, row=0, sticky=tkinter.EW)
            sf.grid_columnconfigure(i + 1, weight=1, uniform='fsb')
            sb.bind('<ButtonPress>', self.try_event(focus(sb)))
        wssf.pack(fill=tkinter.X)
        fontfr.grid_configure(row=1, column=1)
        self.s_families.bind(
            '<<ListboxSelect>>', self.try_event(self.set_score_font_family))

    def create_game_frame(self):
        """Create sample game for demostrating fonts and colours."""
        self.game = game.Game(
            self.chooser,
            ui=self.ui,
            boardfont=self.boardfont,
            tags_variations_comments_font=self.tags_variations_comments_font,
            moves_played_in_game_font=self.moves_played_in_game_font)

        def focus(event=None):
            self.game.score.focus_set()
            
        self.game.score.bind('<ButtonPress>', self.try_event(focus))
        self.game.pgn.get_first_game(''.join((
            '[Event"National Club: Gosport - Wood Green"]',
            '[Site"Gosport"]',
            '[Date"1989-05-07"]',
            '[Round"QFinal"]',
            '[White"Sowray P J"]',
            '[Black"Marsh R"]',
            '[Result"1-0"]',
            'e4c6d4d5exd5cxd5c4Nf6c5(Nf3)(Nc3)e6Nc3b6b4a5Bf4axb4Nb5Na6Qa4Bd7',
            'Bc7Nxc5Qd1Qc8dxc5bxc5Nf3Qb7Nd6Bxd6Bxd6Qb6Be5Ke7Be2Ne4O-Of6Bb2Nc3',
            '{A comment\non two lines}',
            'Bxc3bxc3Qd3(Qb3)Ra3Rfb1Qa7Qc2g6Rb3d4Bc4Rxb3Bxb3Qa6a4Rb8a5e5Bd5Rb2',
            'Qe4Bf5Qh4Qd3(c2(g5)Nd2Qxa5Rxa5Rb1Nf1Rxf1Kxf1Bd3)g4Rb1Rxb1Qxb1Kg2',
            'Kd6Qxf6Kxd5Qxe5Kc6gxf5Qxf5Qe8Kc7Qe7Kc8Ne5c2Qxc5Kd8Qxd4Ke8Qe3Kf8',
            'Kg3Qc8Nd3Kg8f4Qc6Nc1Qa4Qb31-0',
            '\n')))
        self.game.set_game()
        self.game.get_top_widget().pack(fill=tkinter.BOTH, expand=1)
        self.game.get_top_widget().pack_propagate(False)

    def get_options(self):
        """Return (olddefaults, newdefaults) colours."""
        cg = game.Game
        cb = board.Board
        olddefaults = {
            constants.LITECOLOR_NAME: cb.litecolor,
            constants.DARKCOLOR_NAME: cb.darkcolor,
            constants.WHITECOLOR_NAME: cb.whitecolor,
            constants.BLACKCOLOR_NAME: cb.blackcolor,
            constants.LINE_COLOR_NAME: cg.l_color,
            constants.MOVE_COLOR_NAME: cg.m_color,
            constants.ALTERNATIVE_MOVE_COLOR_NAME: cg.am_color,
            constants.VARIATION_COLOR_NAME: cg.v_color,
            }
        for fn in options.font_names:
            olddefaults[fn] = dict()
        fonts.copy_font_attributes(
            tkinter.font.nametofont(constants.MOVES_PLAYED_IN_GAME_FONT).actual(),
            olddefaults[constants.MOVES_PLAYED_IN_GAME_FONT])
        fonts.copy_font_attributes(
            tkinter.font.nametofont(constants.TAGS_VARIATIONS_COMMENTS_FONT).actual(),
            olddefaults[constants.TAGS_VARIATIONS_COMMENTS_FONT])
        fonts.copy_board_font_attributes(
            tkinter.font.nametofont(constants.PIECES_ON_BOARD_FONT).actual(),
            olddefaults[constants.PIECES_ON_BOARD_FONT])
        fonts.copy_board_font_attributes(
            tkinter.font.nametofont(constants.WILDPIECES_ON_BOARD_FONT).actual(),
            olddefaults[constants.WILDPIECES_ON_BOARD_FONT])
        newdefaults = {
            constants.LITECOLOR_NAME: self.game.board.litecolor,
            constants.DARKCOLOR_NAME: self.game.board.darkcolor,
            constants.WHITECOLOR_NAME: self.game.board.whitecolor,
            constants.BLACKCOLOR_NAME: self.game.board.blackcolor,
            constants.LINE_COLOR_NAME: self.game.l_color,
            constants.MOVE_COLOR_NAME: self.game.m_color,
            constants.ALTERNATIVE_MOVE_COLOR_NAME: self.game.am_color,
            constants.VARIATION_COLOR_NAME: self.game.v_color,
            }
        for fn in options.font_names:
            newdefaults[fn] = dict()
        fonts.copy_font_attributes(
            self.tags_variations_comments_font.actual(),
            newdefaults[constants.TAGS_VARIATIONS_COMMENTS_FONT])
        fonts.copy_font_attributes(
            self.moves_played_in_game_font.actual(),
            newdefaults[constants.MOVES_PLAYED_IN_GAME_FONT])
        fonts.copy_board_font_attributes(
            self.boardfont.actual(),
            newdefaults[constants.PIECES_ON_BOARD_FONT])
        fonts.copy_board_font_attributes(
            self.wildpiecesfont.actual(),
            newdefaults[constants.WILDPIECES_ON_BOARD_FONT])
        return olddefaults, newdefaults

    def get_combination(self, option, value):
        tvc = self.tags_variations_comments_font[option] == value
        m = self.moves_played_in_game_font[option] == value
        if tvc:
            if m:
                return 2
            else:
                return 4
        elif m:
            return 1
        else:
            return 3

    def is_ok(self):
        """Returns True if dialogue closed using Ok button."""
        return self.ok

    def on_cancel(self, event=None):
        """Process Cancel button event."""
        self.ok = False
        self.restore_focus.focus_set()
        self.chooser.destroy()

    def on_ok(self, event=None):
        """Process Ok button event."""
        self.ok = True
        self.restore_focus.focus_set()
        self.chooser.destroy()

    def modal_dialogue(self):
        """Display widget as a modal dialogue.

        Activate game navigation so fonts and colours can be seen in game
        context.

        """
        self.restore_focus = self.chooser.focus_get()
        self.chooser.wait_visibility()
        self.chooser.grab_set()
        self.chooser.wait_window()

    def set_board_font_family(self, event=None):
        """Set sample game board family."""
        f = self.b_families
        self.boardfont['family'] = f.get(f.curselection()[0])
        self.game.board.font['family'] = f.get(f.curselection()[0])
        self.b_family['text'] = f.get(f.curselection()[0])

    def set_board_font_weight(self):
        """Set sample game board pieces font."""
        w = {
            0:'normal',
            1:'bold'}
        bfweight = w[self.b_weight.get()]
        self.boardfont['weight'] = bfweight
        self.game.board.font['weight'] = bfweight

    def set_lite_square_bgcolour(self, colour):
        """Set light squares to colour."""
        self.game.board.litecolor = colour
        self.game.board.set_color_scheme()

    def set_dark_square_bgcolour(self, colour):
        """Set dark squares to colour."""
        self.game.board.darkcolor = colour
        self.game.board.set_color_scheme()

    def set_white_piece_fgcolour(self, colour):
        """Set white pieces to colour."""
        self.game.board.whitecolor = colour
        self.game.board.draw_board()

    def set_black_piece_fgcolour(self, colour):
        """Set black pieces to colour."""
        self.game.board.blackcolor = colour
        self.game.board.draw_board()

    def set_tag_rest_of_line(self, colour):
        """Set moves following current move in line to colour."""
        self.game.l_color = colour
        self.game.score.tag_configure('l_color', background=colour)

    def set_tag_move(self, colour):
        """Set current move to colour."""
        self.game.m_color = colour
        self.game.score.tag_configure('m_color', background=colour)

    def set_tag_alternatives(self, colour):
        """Set alternatives to selected next move to colour."""
        self.game.am_color = colour
        self.game.score.tag_configure('am_color', background=colour)

    def set_tag_start_of_line(self, colour):
        """Set moves preceeding current move in line to colour."""
        self.game.v_color = colour
        self.game.score.tag_configure('v_color', background=colour)

    def set_score_font_family(self, event=None):
        """Set game score and variation font family."""
        f = self.s_families
        self.tags_variations_comments_font['family'] = f.get(f.curselection()[0])
        self.moves_played_in_game_font['family'] = f.get(f.curselection()[0])
        self.wildpiecesfont['family'] = f.get(f.curselection()[0])
        self.s_family['text'] = f.get(f.curselection()[0])

    def set_score_font_size(self):
        """Set game score and variation font size to selected size."""
        self.tags_variations_comments_font['size'] = -self.s_size.get()
        self.moves_played_in_game_font['size'] = -self.s_size.get()
    
    def set_score_font_slant(self):
        """Set game score and variation font slants to selected slants."""
        s = {
            4:('italic', 'roman'),
            2:('italic', 'italic'),
            3:('roman', 'roman'),
            1:('roman', 'italic')}
        tvcfslant, mfslant = s[self.s_slant.get()]
        self.tags_variations_comments_font['slant'] = tvcfslant
        self.moves_played_in_game_font['slant'] = mfslant
    
    def set_score_font_weight(self):
        """Set game score and variation font weights to selected weights."""
        w = {
            4:('bold', 'normal'),
            2:('bold', 'bold'),
            3:('normal', 'normal'),
            1:('normal', 'bold')}
        tvcfweight, mfweight = w[self.s_weight.get()]
        self.tags_variations_comments_font['weight'] = tvcfweight
        self.moves_played_in_game_font['weight'] = mfweight
    
    def __del__(self):
        """Force IsOk to return False after destructor call."""
        self.ok = False

    def _make_font_selector(self, frame, title, variable, baserow, command):
        """Create a font selector."""
        tkinter.Label(
            master=frame,
            text=title
            ).grid_configure(column=0, row=baserow)

        def focus(w):
            def fs(e=None):
                w.focus_set()
            return fs
        
        for i, t in (
            (1, 'Moves Played'),
            (2, 'All'),
            (3, 'None'),
            ):
            rb = tkinter.Radiobutton(
                master=frame,
                text=t,
                variable=variable,
                value=i,
                command=self.try_command(command, frame),
                indicatoron=tkinter.FALSE
                )
            rb.grid_configure(column=i, row=baserow, sticky=tkinter.EW)
            rb.bind('<ButtonPress>', self.try_event(focus(rb)))
        rb = tkinter.Radiobutton(
            master=frame,
            text='Tags Variations Comments',
            variable=variable,
            value=4,
            command=self.try_command(command, frame),
            indicatoron=tkinter.FALSE
            )
        rb.grid_configure(
            column=1, row=baserow + 1, columnspan=3, sticky=tkinter.EW)
        rb.bind('<ButtonPress>', self.try_event(focus(rb)))


class ColourChooser(_ColourScheme):

    """Create a colour selection widget.

    Use this widget on screens up to 950 pixels high.
    
    """

    def __init__(self, **ka):
        """Extend to create a widget containing just the colour selectors.

        **ka
        ui is the ChessUI instance
        moves tags and pieces are fonts to be used as basis for modified fonts.
        
        """
        super(ColourChooser, self).__init__(
            height=660,
            title='Chessboard colour chooser',
            **ka)
        self.create_game_frame()
        self.create_colour_frame()
        self.create_buttons(self.get_button_definitions())
        self.modal_dialogue()


class FontChooser(_ColourScheme):

    """Create a font selection widget.

    Use this widget on screens up to 950 pixels high.
    
    """

    def __init__(self, **ka):
        """Extend to create a widget containing just the font selectors.

        **ka
        ui is the ChessUI instance
        moves tags and pieces are fonts to be used as basis for modified fonts.
        
        """
        super(FontChooser, self).__init__(
            height=680,
            title='Chessboard font chooser',
            **ka)
        self.create_font_frame()
        self.create_game_frame()
        self.create_buttons(self.get_button_definitions())
        self.modal_dialogue()


class FontColourChooser(_ColourScheme):

    """Create a font and colour selection widget.

    Use this widget on screens over 950 pixels high.
    
    """

    def __init__(self, **ka):
        """Extend to create a widget containing font and colour selectors.

        **ka
        ui is the ChessUI instance
        moves tags and pieces are fonts to be used as basis for modified fonts.
        
        """
        super(FontColourChooser, self).__init__(
            height=950,
            title='Chessboard style chooser',
            **ka)
        self.create_font_frame()
        self.create_game_frame()
        self.create_colour_frame()
        self.create_buttons(self.get_button_definitions())
        self.modal_dialogue()


class ChessColourSlider(ColourSlider):

    """Create colour selection widget containing Red Green and Blue sliders.
    
    """

    def __init__(
        self,
        master=None,
        row=None,
        label='',
        resolution=2,
        colour='grey',
        TagConfigureBGcolour=None):
        """Extend with callback to attach a colour to a Text tag."""
        super(ChessColourSlider, self).__init__(
            master=master,
            row=row,
            label=label,
            resolution=resolution,
            colour=colour)
        self.TagConfigureBGcolour = TagConfigureBGcolour

    def delta_red_colour(self, event=None):
        """Extend to attach colour to Text tag after change to Red."""
        super(ChessColourSlider, self).delta_red_colour(event=event)
        self.TagConfigureBGcolour(self.get_colour())

    def set_red_colour(self, event=None):
        """Extend to attach colour to Text tag after setting Red."""
        super(ChessColourSlider, self).set_red_colour(event=event)
        self.TagConfigureBGcolour(self.get_colour())

    def delta_green_colour(self, event=None):
        """Extend to attach colour to Text tag after change to Green."""
        super(ChessColourSlider, self).delta_green_colour(event=event)
        self.TagConfigureBGcolour(self.get_colour())

    def set_green_colour(self, event=None):
        """Extend to attach colour to Text tag after setting Green."""
        super(ChessColourSlider, self).set_green_colour(event=event)
        self.TagConfigureBGcolour(self.get_colour())

    def delta_blue_colour(self, event=None):
        """Extend to attach colour to Text tag after change to Blue."""
        super(ChessColourSlider, self).delta_blue_colour(event=event)
        self.TagConfigureBGcolour(self.get_colour())

    def set_blue_colour(self, event=None):
        """Extend to attach colour to Text tag after setting Blue."""
        super(ChessColourSlider, self).set_blue_colour(event=event)
        self.TagConfigureBGcolour(self.get_colour())




# querydisplay.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widgets to display and edit game selection rules.

These four classes display game selection rules in the main window: they are
used in the querygrid module.

The QueryDisplay class binds events to navigate between widgets.

The DatabaseQueryDisplay class adds delete record to the QueryDisplay
class.

The DatabaseQueryInsert class adds insert record to the
DatabaseQueryDisplay class but does not bind delete record to any events.

The DatabaseQueryEdit class adds edit record to the DatabaseQueryInsert
class but does not bind delete record to any events.

These three classes display games in their own Toplevel widget: they are used
in the gamedbdelete, gamedbedit, and gamedbshow, modules.

The QueryDialogue class binds events to navigate between widgets.

The DialogueQueryDisplay class adds insert and delete record to the
QueryDialogue class.

The DialogueQueryEdit class adds insert and edit record to the
QueryDialogue class.

"""

import tkinter
import tkinter.messagebox

from solentware_misc.workarounds import dialogues
from solentware_base.core.where import WhereError

from solentware_grid.gui.dataedit import RecordEdit
from solentware_grid.gui.datadelete import RecordDelete
from solentware_grid.core.dataclient import DataNotify

from ..core.querystatement import QueryStatement
from .query import Query
from .queryedit import QueryEdit
from ..core.chessrecord import ChessDBrecordQuery
from .chessexception import ChessException
from .eventspec import EventSpec
from .display import Display
from .. import APPLICATION_NAME


class QueryDisplay(ChessException, Display):
    
    """Manage UI interaction with database for displayed game selection rules.

    QueryDisplay is a subclass of DataNotify so that modifications to
    the database record outside an instance prevent database update using
    the instance.  This class provides methods to update the database from
    the instance; and to switch to other displayed game selection rules.

    Subclasses provide the widget to display the game selection rules.
    
    """

    binding_labels = tuple(
        [b[1:] for b in (
            EventSpec.selectiondisplay_to_position_grid,
            EventSpec.selectiondisplay_to_active_game,
            EventSpec.selectiondisplay_to_game_grid,
            EventSpec.selectiondisplay_to_repertoire_grid,
            EventSpec.selectiondisplay_to_active_repertoire,
            EventSpec.selectiondisplay_to_repertoire_game_grid,
            EventSpec.selectiondisplay_to_partial_grid,
            EventSpec.selectiondisplay_to_selection_rule_grid,
            EventSpec.selectiondisplay_to_previous_selection,
            EventSpec.selectiondisplay_to_next_selection,
            EventSpec.selectiondisplay_to_partial_game_grid,
            EventSpec.tab_traverse_backward,
            EventSpec.tab_traverse_forward,
            )])

    def __init__(self, sourceobject=None, **ka):
        """Extend and link game selection rule to database.

        sourceobject - link to database.

        """
        super(QueryDisplay, self).__init__(**ka)
        self.blockchange = False
        if self.ui.base_selections.datasource:
            self.set_data_source(self.ui.base_selections.get_data_source())
        self.sourceobject = sourceobject
        self.insertonly = sourceobject is None
        self.recalculate_after_edit = sourceobject

    def _bind_for_board_navigation(self):
        """Set bindings to navigate game selection rule on pointer click."""
        self.bind_score_pointer_for_board_navigation(True)

    def bind_for_widget_navigation(self):
        """Set bindings to give focus to this selection rule on pointer click.
        """
        self.bind_score_pointer_for_widget_navigation(True)

    def bind_off(self):
        """Disable all bindings."""
        
        # Replicate structure of __bind_on for deleting bindings.
        for sequence, function in (
            (EventSpec.selectiondisplay_to_partial_grid, ''),
            (EventSpec.selectiondisplay_to_selection_rule_grid, ''),
            (EventSpec.selectiondisplay_to_previous_selection, ''),
            (EventSpec.selectiondisplay_to_next_selection, ''),
            (EventSpec.selectiondisplay_to_selection_game_grid, ''),
            (EventSpec.selectiondisplay_to_repertoire_grid, ''),
            (EventSpec.selectiondisplay_to_active_repertoire, ''),
            (EventSpec.selectiondisplay_to_repertoire_game_grid, ''),
            (EventSpec.selectiondisplay_to_position_grid, ''),
            (EventSpec.selectiondisplay_to_active_game, ''),
            (EventSpec.selectiondisplay_to_game_grid, ''),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    # This method exists because the bindings cannot be done until all __init__
    # methods in superclasses have been called.
    def initialize_bindings(self):
        """Initialize all bindings."""
        self.__bind_on()
        for sequence, function in (
            (EventSpec.tab_traverse_forward,
             self.traverse_forward),
            (EventSpec.tab_traverse_backward,
             self.traverse_backward),
            (EventSpec.tab_traverse_round,
             self.traverse_round),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def bind_on(self):
        """Enable all bindings."""
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""

        # Same bindings in initialize_bindings() and bind_on() in this class.
        for sequence, function in (
            ('<Control-ButtonPress-1>', ''),
            ('<ButtonPress-1>', self.try_event(self.give_focus_to_widget)),
            ('<ButtonPress-3>', self.try_event(self.popup_inactive_menu)),
            ):
            self.score.bind(sequence, function)
        for sequence, function in (
            (EventSpec.selectiondisplay_to_partial_grid,
             self.set_focus_partial_grid),
            (EventSpec.selectiondisplay_to_selection_rule_grid,
             self.set_focus_selection_rule_grid),
            (EventSpec.selectiondisplay_to_previous_selection,
             self.prior_item),
            (EventSpec.selectiondisplay_to_next_selection,
             self.next_item),
            (EventSpec.selectiondisplay_to_partial_game_grid,
             self.set_focus_partial_game_grid),
            (EventSpec.selectiondisplay_to_repertoire_grid,
             self.set_focus_repertoire_grid),
            (EventSpec.selectiondisplay_to_active_repertoire,
             self.set_focus_repertoirepanel_item),
            (EventSpec.selectiondisplay_to_repertoire_game_grid,
             self.set_focus_repertoire_game_grid),
            (EventSpec.selectiondisplay_to_position_grid,
             self.set_focus_position_grid),
            (EventSpec.selectiondisplay_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.selectiondisplay_to_game_grid,
             self.set_focus_game_grid),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def _cycle_item(self, prior=False):
        """Select next game selection rule on display."""
        items = self.ui.selection_items
        losefocus = items.active_item
        losefocus.bind_for_widget_navigation()
        items.cycle_active_item(prior=prior)
        self.ui.configure_selection_grid()
        gainfocus = items.active_item
        gainfocus.refresh_game_list()
        gainfocus._bind_for_board_navigation()
        gainfocus.takefocus_widget.focus_set()
        gainfocus.set_statusbar_text()

    def give_focus_to_widget(self, event=None):
        """Select game selection rule on display by mouse click."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        losefocus, gainfocus = self.ui.selection_items.give_focus_to_widget(
            event.widget)
        if losefocus is not gainfocus:
            self.ui.configure_selection_grid()
            self.score.after(
                0,
                func=self.try_command(self.ui._set_selection_name, self.score))
            self.score.after(
                0,
                func=self.try_command(gainfocus.refresh_game_list, self.score))
        return 'break'

    def delete_item_view(self, event=None):
        """Remove game selection rule item from screen."""
        self.ui.delete_selection_rule_view(self)

    def insert_selection_rule_database(self, event=None):
        """Add game selection rule to database."""
        if self.ui.selection_items.active_item is None:
            dialogues.showinfo(
                title='Insert Game Selection Rule',
                message='No active game selection rule to insert into database.')
            return

        # This should see if game selection rule with same name already exists,
        # after checking for database open, and offer option to insert anyway.
        if self.ui.database is None:
            dialogues.showinfo(
                title='Insert Game Selection Rule',
                message='Cannot add game selection rule:\n\nNo database open.')
            return

        datasource = self.ui.base_selections.get_data_source()
        if datasource is None:
            dialogues.showinfo(
                title='Insert Game Selection Rule',
                message='Cannot add game selection rule:\n\nRule list hidden.')
            return
        updater = ChessDBrecordQuery()
        updater.value.set_database(datasource.dbhome)
        updater.value.dbset = self.ui.base_games.datasource.dbset
        uvpqs = updater.value.process_query_statement(
            self.get_name_query_statement_text())
        if not len(updater.value.get_name_text()):
            dialogues.showerror(
                title='Insert Game Selection Rule',
                message=''.join((
                    "The selection rule has no name.\n\nPlease enter it's ",
                    "name as the first line of text.'")))
            return
        if not uvpqs:
            if not updater.value.where_error:
                if tkinter.messagebox.YES != dialogues.askquestion(
                    title='Insert Game Selection Rule',
                    message=''.join((
                        'Confirm request to add game selection rule named:\n\n',
                        updater.value.get_name_text(),
                        '\n\nto database.\n\n',
                        'Note validation of the statement failed but no ',
                        'information is available.',
                        ))):
                    dialogues.showinfo(
                        title='Insert Game Selection Rule',
                        message=''.join(('Add game selection rule to ',
                                         'database abandonned.')))
                    return
            elif tkinter.messagebox.YES != dialogues.askquestion(
                title='Insert Game Selection Rule',
                message=''.join((
                    'Confirm request to add game selection rule named:\n\n',
                    updater.value.get_name_text(),
                    '\n\nto database.\n\n',
                    updater.value.where_error.get_error_report(
                        self.ui.base_games.get_data_source())))):
                dialogues.showinfo(
                    title='Insert Game Selection Rule',
                    message=''.join(('Add game selection rule to ',
                                     'database abandonned.')))
                return
        elif tkinter.messagebox.YES != dialogues.askquestion(
            title='Insert Game Selection Rule',
            message=''.join((
                'Confirm request to add game selection rule named:\n\n',
                updater.value.get_name_text(),
                '\n\nto database.',))):
            dialogues.showinfo(
                title='Insert Game Selection Rule',
                message='Add game selection rule to database abandonned.')
            return
        editor = RecordEdit(updater, None)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        updater.key.recno = None#0
        editor.put()
        dialogues.showinfo(
            title='Insert Game Selection Rule',
            message=''.join(('Game selection rule "',
                             updater.value.get_name_text(),
                             '" added to database.')))
        
    def next_item(self, event=None):
        """Select next game selection rule display.

        Call _next_selection_rule after 1 millisecond to allow message display

        """
        if self.ui.selection_items.count_items_in_stack() > 1:
            self.ui._set_find_selection_name_games(0)
            self.score.after(
                1, func=self.try_command(self._next_selection_rule, self.score))

    def _next_selection_rule(self):
        """Generate next game selection rule display.

        Call from next_item only.

        """
        self._cycle_item(prior=False)

    # on_game_change() is not called, but do nothing if it is.
    def on_game_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if self.sourceobject is not None:
            pass

    def on_selection_change(self, instance):
        """Prevent update from self if instance refers to same record.

        The list of games is updated by the 'List Games' popup menu option,
        but not as a consequence of updating the database.
        """
        if instance.newrecord:

            # Editing an existing record.
            value = instance.newrecord.value
            key = instance.newrecord.key

        else:

            # Inserting a new record.
            value = instance.value
            key = instance.key

        if self.sourceobject is not None:
            if (key == self.sourceobject.key and
                self.datasource.dbname == self.sourceobject.dbname and
                self.datasource.dbset == self.sourceobject.dbset):
                self.blockchange = True

    def prior_item(self, event=None):
        """Select previous game selection rule display.

        Call _prior_selection_rule after 1 millisecond to allow message display

        """
        if self.ui.selection_items.count_items_in_stack() > 1:
            self.ui._set_find_selection_name_games(-2)
            self.score.after(
                1, func=self.try_command(
                    self._prior_selection_rule, self.score))

    def _prior_selection_rule(self):
        """Generate previous game selection rule display.

        Call from prior_item only.

        """
        self._cycle_item(prior=True)

    def set_insert_or_delete(self):
        """Convert edit display to insert display.

        Selection rules displayed for editing from a database are not closed
        if the database is closed.  They are converted to insert displays and
        can be used to add new records to databases opened later.
        
        """
        self.sourceobject = None

    def get_text_for_statusbar(self):
        """"""
        return ''.join(
            ('Please wait while finding games for game selection rule ',
             self.query_statement.get_name_text(),
             ))

    def get_selection_text_for_statusbar(self):
        """"""
        return self.query_statement.get_name_text()
        
    def bind_toplevel_navigation(self):
        """Set bindings for popup menu for QueryDisplay instance."""
        navigation_map = {
            EventSpec.selectiondisplay_to_position_grid[1]:
            self.set_focus_position_grid,
            EventSpec.selectiondisplay_to_active_game[1]:
            self.set_focus_gamepanel_item_command,
            EventSpec.selectiondisplay_to_game_grid[1]:
            self.set_focus_game_grid,
            EventSpec.selectiondisplay_to_repertoire_grid[1]:
            self.set_focus_repertoire_grid,
            EventSpec.selectiondisplay_to_active_repertoire[1]:
            self.set_focus_repertoirepanel_item_command,
            EventSpec.selectiondisplay_to_partial_grid[1]:
            self.set_focus_partial_grid,
            EventSpec.selectiondisplay_to_active_partial[1]:
            self.set_focus_partialpanel_item_command,
            EventSpec.selectiondisplay_to_repertoire_game_grid[1]:
            self.set_focus_repertoire_game_grid,
            EventSpec.selectiondisplay_to_partial_grid[1]:
            self.set_focus_partial_grid,
            EventSpec.selectiondisplay_to_partial_game_grid[1]:
            self.set_focus_partial_game_grid,
            EventSpec.tab_traverse_backward[1]:
            self.traverse_backward,
            EventSpec.tab_traverse_forward[1]:
            self.traverse_forward,
            }
        for nm, widget in (
            ({EventSpec.selectiondisplay_to_previous_selection[1]:
              self.prior_item,
              EventSpec.selectiondisplay_to_next_selection[1]: self.next_item,
              },
             self),
            ):
            nm.update(navigation_map)
            widget.add_navigation_to_viewmode_popup(
                bindings=nm, order=self.binding_labels)

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        self.ui.give_focus_backward(self.ui.selection_items)
        return 'break'

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        self.ui.give_focus_forward(self.ui.selection_items)
        return 'break'

    def traverse_round(self, event=None):
        """Give focus to next widget within active item in traversal order."""
        return 'break'

    def process_and_set_selection_rule_list(self, event=None):
        """Display games matching edited game selection rule."""
        if self.ui.base_games.datasource:
            self.query_statement.set_database(
                self.ui.base_games.datasource.dbhome)
        try:
            self.query_statement.process_query_statement(
                self.score.get('1.0', tkinter.END))
        except WhereError as exc:
            dialogues.showinfo(
                title=' '.join(('Game Selection Rule Error')),
                message=str(exc))
            return 'break'
        finally:
            self.query_statement.set_database()
        self.refresh_game_list()
        return 'break'


class DatabaseQueryDisplay(QueryDisplay, Query, DataNotify):
    
    """Display game selection rule from database and allow delete and insert.

    Query provides the widget and QueryDisplay the database interface.
    
    """

    def __init__(self, **ka):
        """Create display widget and display game selection rule from database.

        See superclasses for argument descriptions.

        """
        super(DatabaseQueryDisplay, self).__init__(**ka)
        self.initialize_bindings()

        # Copied from QueryEdit class in queryedit module on assumption the
        # code should be here too.
        # It is needed if a query is displayed by F11, then the database is
        # closed and it, or anothr one, is opened.
        self.selection_token_checker = QueryStatement()
        if self.ui.base_games.datasource:
            self.selection_token_checker.dbset = self.ui.base_games.datasource.dbset

    def bind_off(self):
        """Disable all bindings."""
        super(DatabaseQueryDisplay, self).bind_off()
        for sequence, function in (
            (EventSpec.databaseselectiondisplay_show_game_list, ''),
            (EventSpec.databaseselectiondisplay_insert, ''),
            (EventSpec.databaseselectiondisplay_delete, ''),
            (EventSpec.databaseselectiondisplay_dismiss, ''),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    # This method exists because the bindings cannot be done until all __init__
    # methods in superclasses have been called.
    def initialize_bindings(self):
        """Initialize all bindings."""
        super(DatabaseQueryDisplay, self).initialize_bindings()

        # Here because superclass order is QueryDisplay, Query.
        self.inactive_popup = tkinter.Menu(master=self.score, tearoff=False)

        for function, accelerator in (
            (self.set_focus_panel_item_command,
             EventSpec.databaseselectiondisplay_make_active),
            (self.delete_item_view,
             EventSpec.databaseselectiondisplay_dismiss_inactive),
            ):
            self.inactive_popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.inactive_popup),
                accelerator=accelerator[2])
        self.__bind_on()
        for function, accelerator in (
            (self.process_and_set_selection_rule_list,
             EventSpec.databaseselectiondisplay_show_game_list),
            (self.delete_item_view,
             EventSpec.databaseselectiondisplay_dismiss),
            ):
            self.viewmode_popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.viewmode_popup),
                accelerator=accelerator[2])
        for function, accelerator in (
            (self.insert_selection_rule_database,
             EventSpec.databaseselectiondisplay_insert),
            (self.delete_selection_rule_database,
             EventSpec.databaseselectiondisplay_delete),
            ):
            self.viewmode_database_popup.add_command(
                label=accelerator[1],
                command=self.try_command(
                    function, self.viewmode_database_popup),
                accelerator=accelerator[2])
        self.bind_toplevel_navigation()

    def bind_on(self):
        """Enable all bindings."""
        super(DatabaseQueryDisplay, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Common to bind_on() and initialize_bindings()"""
        for sequence, function in (
            (EventSpec.databaseselectiondisplay_show_game_list,
             self.process_and_set_selection_rule_list),
            (EventSpec.databaseselectiondisplay_insert,
             self.insert_selection_rule_database),
            (EventSpec.databaseselectiondisplay_delete,
             self.delete_selection_rule_database),
            (EventSpec.databaseselectiondisplay_dismiss,
             self.delete_item_view),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def delete_selection_rule_database(self, event=None):
        """Remove game selection rule from database."""
        if self.ui.database is None:
            dialogues.showinfo(
                title='Delete Game Selection Rule',
                message=''.join(('Cannot delete game selection rule:\n\n',
                                 'No database open.')))
            return
        datasource = self.ui.base_selections.get_data_source()
        if datasource is None:
            dialogues.showerror(
                title='Delete Game Selection Rule',
                message=''.join(('Cannot delete game selection rule:\n\n',
                                 'Rule list hidden.')))
            return
        if self.sourceobject is None:
            dialogues.showinfo(
                title='Delete Game Selection Rule',
                message=''.join(('The game selection rule to delete has not ',
                                 'been given.\n\nProbably because database ',
                                 'has been closed and opened since this copy ',
                                 'was displayed.')))
            return
        if self.blockchange:
            dialogues.showinfo(
                title='Delete Game Selection Rule',
                message='\n'.join((
                    'Cannot delete game selection rule.',
                    'Record has been amended since this copy displayed.')))
            return
        if tkinter.messagebox.YES != dialogues.askquestion(
            title='Delete Game Selection Rule',
            message='Confirm request to delete game selection rule.'):
            return
        s = self.query_statement
        if s.where_error is not None:
            v = self.sourceobject.value
            if (s.get_name_text() != v.get_name_text() or
                s.where_error != v.where_error or
                s.get_query_statement_text() != v.get_query_statement_text()):
                dialogues.showinfo(
                    title='Delete Game Selection Rule',
                    message='\n'.join((
                        'Cannot delete game selection rule.',
                        ' '.join((
                            'Rule on display is not same as',
                            'rule from record.')))))
                return
        editor = RecordDelete(self.sourceobject)
        editor.set_data_source(datasource, editor.on_data_change)
        editor.delete()
        dialogues.showinfo(
            title='Delete Game Selection Rule',
            message=''.join(('Game selection rule "',
                             self.sourceobject.value.get_name_text(),
                             '" deleted from database.')))


class DatabaseQueryInsert(QueryDisplay, QueryEdit, DataNotify):
    
    """Display game selection rule from database allowing insert.

    QueryEdit provides the widget and QueryDisplay the database
    interface.
    
    """

    def __init__(self, **ka):
        """Create editor widget and display game selection rule from database.

        See superclasses for argument descriptions.

        """
        super().__init__(**ka)
        self.initialize_bindings()

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        for sequence, function in (
            (EventSpec.databaseselectionedit_show_game_list, ''),
            (EventSpec.databaseselectionedit_insert, ''),
            (EventSpec.databaseselectionedit_dismiss, ''),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    # This method exists because the bindings cannot be done until all __init__
    # methods in superclasses have been called.
    def initialize_bindings(self):
        """Initialize all bindings."""
        super().initialize_bindings()

        # Here because superclass order is QueryDisplay, QueryEdit.
        self.inactive_popup = tkinter.Menu(master=self.score, tearoff=False)

        for function, accelerator in (
            (self.set_focus_panel_item_command,
             EventSpec.databaseselectionedit_make_active),
            (self.delete_item_view,
             EventSpec.databaseselectionedit_dismiss_inactive),
            ):
            self.inactive_popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.inactive_popup),
                accelerator=accelerator[2])
        self.__bind_on()
        for function, accelerator in (
            (self.process_and_set_selection_rule_list,
             EventSpec.databaseselectionedit_show_game_list),
            (self.delete_item_view,
             EventSpec.databaseselectionedit_dismiss),
            ):
            self.viewmode_popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.viewmode_popup),
                accelerator=accelerator[2])
        for function, accelerator in (
            (self.insert_selection_rule_database,
             EventSpec.databaseselectionedit_insert),
            ):
            self.viewmode_database_popup.add_command(
                label=accelerator[1],
                command=self.try_command(
                    function, self.viewmode_database_popup),
                accelerator=accelerator[2])
        self.bind_toplevel_navigation()

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Common to bind_on() and initialize_bindings()"""
        for sequence, function in (
            (EventSpec.databaseselectionedit_show_game_list,
             self.process_and_set_selection_rule_list),
            (EventSpec.databaseselectionedit_insert,
             self.insert_selection_rule_database),
            (EventSpec.databaseselectionedit_dismiss,
             self.delete_item_view),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def insert_char_to_right(self, char):
        """"""
        r = super().insert_char_to_right(char)
        if r is None:
            return None
        elif r:
            return True
        else:
            self.recalculate_after_edit = None
            return None

    def delete_char_left(self, event):
        """"""
        if not self.current:
            return 'break'
        self.recalculate_after_edit = None
        return super().delete_char_left(event)

    def delete_char_right(self, event):
        """"""
        if not self.current:
            return 'break'
        self.recalculate_after_edit = None
        return super().delete_char_right(event)


class DatabaseQueryEdit(DatabaseQueryInsert):
    
    """Display game selection rule from database allowing edit and insert.

    DatabaseQueryEdit adds insert to the database interface.
    
    """

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        for sequence, function in (
            (EventSpec.databaseselectionedit_update, ''),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    # This method exists because the bindings cannot be done until all __init__
    # methods in superclasses have been called.
    def initialize_bindings(self):
        """Initialize all bindings."""
        super().initialize_bindings()
        self.__bind_on()
        for function, accelerator in (
            (self.update_selection_rule_database,
             EventSpec.databaseselectionedit_update),
            ):
            self.viewmode_database_popup.add_command(
                label=accelerator[1],
                command=self.try_command(
                    function, self.viewmode_database_popup),
                accelerator=accelerator[2])

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Common to bind_on() and initialize_bindings()"""
        for sequence, function in (
            (EventSpec.databaseselectionedit_update,
             self.update_selection_rule_database),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def update_selection_rule_database(self, event=None):
        """Modify existing game selection rule record."""
        if self.ui.database is None:
            dialogues.showinfo(
                title='Edit Game Selection Rule',
                message='Cannot edit repertoire:\n\nNo database open.')
            return
        datasource = self.ui.base_selections.get_data_source()
        if datasource is None:
            dialogues.showinfo(
                title='Edit Game Selection Rule',
                message='Cannot edit game selection rule:\n\nRule list hidden.')
            return
        if self.sourceobject is None:
            dialogues.showinfo(
                title='Edit Game Selection Rule',
                message=''.join(('The game selection rule to edit has not ',
                                 'been given.\n\nProbably because database ',
                                 'has been closed and opened since this copy ',
                                 'was displayed.')))
            return
        if self.blockchange:
            dialogues.showinfo(
                title='Edit Game Selection Rule',
                message='\n'.join((
                    'Cannot edit game selection rule.',
                    'Rule has been amended since this copy was displayed.')))
            return
        original = ChessDBrecordQuery()
        original.set_database(datasource.dbhome)
        original.load_record(
            (self.sourceobject.key.recno,
             self.sourceobject.srvalue))

        # is it better to use DataClient directly?
        # Then original would not be used. Instead DataSource.new_row
        # gets record keyed by sourceobject and update is used to edit this.
        updater = ChessDBrecordQuery()
        updater.value.set_database(datasource.dbhome)
        updater.value.dbset = self.ui.base_games.datasource.dbset
        uvpqs = updater.value.process_query_statement(
            self.get_name_query_statement_text())
        if not len(updater.value.get_name_text()):
            dialogues.showerror(
                title='Edit Game Selection Rule',
                message=''.join((
                    "The selection rule has no name.\n\nPlease enter it's ",
                    "name as the first line of text.'")))
            return
        if not uvpqs:
            if not updater.value.where_error:
                if tkinter.messagebox.YES != dialogues.askquestion(
                    title='Edit Game Selection Rule',
                    message=''.join((
                        'Confirm request to edit game selection rule ',
                        'named:\n\n',
                        updater.value.get_name_text(),
                        '\n\nto database.\n\n',
                        'Note validation of the statement failed but no ',
                        'information is available.',
                        ))):
                    dialogues.showinfo(
                        title='Edit Game Selection Rule',
                        message=''.join(('Edit game selection rule to ',
                                         'database abandonned.')))
                    return
            elif tkinter.messagebox.YES != dialogues.askquestion(
                title='Edit Game Selection Rule',
                message=''.join((
                    'Confirm request to edit game selection rule named:\n\n',
                    updater.value.get_name_text(),
                    '\n\nto database.\n\n',
                    updater.value.where_error.get_error_report(
                        self.ui.base_games.get_data_source())))):
                dialogues.showinfo(
                    title='Edit Game Selection Rule',
                    message=''.join(('Edit game selection rule to ',
                                     'database abandonned.')))
                return
        elif tkinter.messagebox.YES != dialogues.askquestion(
            title='Edit Game Selection Rule',
            message=''.join((
                'Confirm request to edit game selection rule named:\n\n',
                updater.value.get_name_text(),
                '\n\non database.',))):
            dialogues.showinfo(
                title='Edit Game Selection Rule',
                message='Edit game selection rule to database abandonned.')
            return
        editor = RecordEdit(updater, original)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        original.set_database(editor.get_data_source().dbhome)
        updater.key.recno = original.key.recno
        editor.edit()
        dialogues.showinfo(
            title='Edit Game Selection Rule',
            message=''.join(('Game selection rule "',
                             updater.value.get_name_text(),
                             '" amended on database.')))


class QueryDialogue(ChessException):
    
    """Manage UI interaction with database for a displayed game selection rule.

    Subclasses provide the widget to display the game selection rule.
    
    """

    # Formally the same as GameDialogue in gamedisplay module.
    # Follow whatever style of assignment is used there if this binding_labels
    # needs to be non-empty.
    binding_labels = tuple()

    def initialize_bindings(self):
        """Enable all bindings."""
        self.bind_score_pointer_for_board_navigation(True)
        self.bind_toplevel_navigation()
        
    def bind_toplevel_navigation(self):
        """Set bindings for popup menu for QueryDialogue instance."""

        # Formally the same as GameDialogue in gamedisplay module.
        # The effect is add no bindings to an empty bindings map.
        navigation_map = {}
        for nm, widget in (
            ({},
             self),
            ):
            nm.update(navigation_map)
            widget.add_navigation_to_viewmode_popup(
                bindings=nm, order=self.binding_labels)


class DialogueQueryDisplay(Query, QueryDialogue):
    
    """Display a game selection rule from a database allowing delete and insert.
    """

    def __init__(self, **ka):
        """Extend and link game selection rule to database.
        """
        super(DialogueQueryDisplay, self).__init__(**ka)
        self.initialize_bindings()


class DialogueQueryEdit(QueryEdit, QueryDialogue):
    
    """Display a game selection rule from a database allowing edit and insert.
    """

    def __init__(self, **ka):
        """Extend and link game selection rule to database.

        """
        super(DialogueQueryEdit, self).__init__(**ka)
        self.initialize_bindings()
